#!/usr/bin/env python
# 

from __future__ import print_function

import os, sys, re, json, time, astropy, copy
import numpy as np
from astropy.table import Table, Column, hstack
from numpy import log10, power as pow

#from astropy.cosmology import FlatLambdaCDM
##cosmo = FlatLambdaCDM(H0=73, Om0=0.27, Tcmb0=2.725) # replaced since 2019-02-21
#cosmo = FlatLambdaCDM(H0=70, Om0=0.27, Tcmb0=2.725)
if not (os.path.dirname(os.path.abspath(__file__)) in sys.path): sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import apply_cosmology
cosmo = apply_cosmology.cosmo

if sys.version_info.major >= 3:
    long = int
else:
    pass



# 
# def 
# 
def calc_gas_fraction_Scoville2017(z, lgMstar=10.5, DeltaMS=0.0):
    # Scoville et al. 2017 (The Astrophysical Journal, 837:150 (20pp), 2017 March 10)
    # Table 2
    Ratio_M_molgas_M_star = 0.71 * (1+z)**1.84 * (10**DeltaMS)**0.32 * (10**(lgMstar-10.0))**(-0.70)
    return Ratio_M_molgas_M_star

def calc_gas_fraction_Tacconi2018_beta_2(z, lgMstar=10.5, DeltaMS=0.0):
    # Tacconi et al. 2018 (The Astrophysical Journal, 853:179 (22pp), 2018 February 1 https://doi.org/10.3847/1538-4357/aaa4b4)
    # Table 3
    # Best^c beta=2 with S14 (Speagle+2014) MS
    A = +0.12
    B = -3.62
    F = +0.66
    beta = 2.0
    C = +0.53
    D = -0.35
    E = +0.11
    DeltaRe = 0.0 # log10 (R_e / R_e_0)
    log10_Ratio_M_molgas_M_star = A + B * (np.log10(1+z)-F)**beta + C * DeltaMS + D * (lgMstar-10.7) + E * (DeltaRe)
    Ratio_M_molgas_M_star = 10**log10_Ratio_M_molgas_M_star
    return Ratio_M_molgas_M_star

def calc_gas_fraction_Tacconi2018(z, lgMstar=10.5, DeltaMS=0.0):
    return calc_gas_fraction_Tacconi2018_beta_2(z, lgMstar, DeltaMS)

def calc_gas_fraction_A3COSMOS(cosmic_age=None, z=None, lgMstar=10.5, DeltaMS=0.0):
    # A3COSMOS 
    # Sample 20181203 
    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plot/Plot_z_deltaGas_data_fitting/20190715_with_gas_mass_calibration_H17/fitting_dzliu_function/best_fit_via_MCMC.json'
    #popt = [0.38380575488270824, -0.8341369033600499, -0.15468695242992148, 1.0834290154064226, 0.10008727595632205, 0.040159361310305286]
    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plot/Plot_z_deltaGas_data_fitting/20190802_with_gas_mass_calibration_H17_with_own_uncertainty/fitting_dzliu_function/best_fit_via_MCMC.json'
    #popt = [0.4993149209610621, -0.665870841347096, -0.1488184998332125, 0.8970477945369266, 0.0989348364520346, 0.02889191643165745]
    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plot/Plot_z_deltaGas_data_fitting/20190827_with_gas_mass_calibration_H17/fitting_dzliu_function/best_fit_via_MCMC.json' # -- now with Cortzen+2019
    popt = [0.4195381171312362, -0.6906526862112656, -0.1542557899746253, 0.9339052286147904, 0.11949845461785102, 0.03204158046196248]
    a, b, c, d, ak, ck = popt
    if z is None and cosmic_age is None:
        print('Error! Please input either z or cosmic_age')
        sys.exit()
    else:
        if cosmic_age is not None:
            cosmoAge = cosmic_age
        else:
            cosmoAge = cosmo.age(z).value
    log10_Ratio_M_molgas_M_star = (a+ak*(lgMstar-10.0))*DeltaMS + b*(lgMstar-10.0) + (c+ck*(lgMstar-10.0))*(cosmoAge) + d
    Ratio_M_molgas_M_star = 10**log10_Ratio_M_molgas_M_star
    return Ratio_M_molgas_M_star

def calc_gas_fraction_A3COSMOS_with_Leslie2019_MS(cosmic_age=None, z=None, lgMstar=10.5, DeltaMS=0.0):
    # A3COSMOS 
    # Sample 20181203 
    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plot/Plot_z_deltaGas_data_fitting/20190827_with_gas_mass_calibration_H17_and_main_sequence_Leslie2019/fitting_dzliu_function/best_fit_via_MCMC.json'
    popt = [0.3781725710882265, -0.8360054443744662, -0.1557707273000366, 1.0923711580061433, 0.09940673198827188, 0.03940816266321434]
    a, b, c, d, ak, ck = popt
    if z is None and cosmic_age is None:
        print('Error! Please input either z or cosmic_age')
        sys.exit()
    else:
        if cosmic_age is not None:
            cosmoAge = cosmic_age
        else:
            cosmoAge = cosmo.age(z).value
    log10_Ratio_M_molgas_M_star = (a+ak*(lgMstar-10.0))*DeltaMS + b*(lgMstar-10.0) + (c+ck*(lgMstar-10.0))*(cosmoAge) + d
    Ratio_M_molgas_M_star = 10**log10_Ratio_M_molgas_M_star
    return Ratio_M_molgas_M_star

def calc_gas_fraction_A3COSMOS_with_Scoville2017_MS(cosmic_age=None, z=None, lgMstar=10.5, DeltaMS=0.0):
    # A3COSMOS 
    # Sample 20181203 
    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plot/Plot_z_deltaGas_data_fitting/20190827_with_gas_mass_calibration_H17_and_main_sequence_Scoville2017/fitting_dzliu_function/best_fit_via_MCMC.json'
    popt = [0.42803584038283427, -1.0941775869942276, -0.168209612312487, 1.207999567254241, 0.08377200755981207, 0.06373985587594078]
    a, b, c, d, ak, ck = popt
    if z is None and cosmic_age is None:
        print('Error! Please input either z or cosmic_age')
        sys.exit()
    else:
        if cosmic_age is not None:
            cosmoAge = cosmic_age
        else:
            cosmoAge = cosmo.age(z).value
    log10_Ratio_M_molgas_M_star = (a+ak*(lgMstar-10.0))*DeltaMS + b*(lgMstar-10.0) + (c+ck*(lgMstar-10.0))*(cosmoAge) + d
    Ratio_M_molgas_M_star = 10**log10_Ratio_M_molgas_M_star
    return Ratio_M_molgas_M_star

#def calc_gas_fraction_A3COSMOS(cosmoAge, lgMstar=10.5, DeltaMS=0.0):
#    # A3COSMOS 
#    # Sample 20181203 
#    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plots/Plot_z_deltaGas_data_fitting/best_fit_via_CurveFit.json'
#    #popt = [0.7176359685433024, -0.3304845603259402, 0.008136603197619791, -0.07859177961327789, -0.026455117351337622] #<20190122_with_Leroy_GDR_with_Genzel_Eq12a_MZR_with_KMT09_fmol>#
#    #popt = [0.6829771406895456, -0.5516107864334655, 0.021790228645462928, -0.1558086567222849, 0.752820037618907]
#    #popt = [0.6569, -0.6628, 0.02804, -0.1689, +0.9282] #<20190205_with_RemyRuyer_GDR_with_Genzel_Eq12a_MZR_with_KMT09_fmol>#
#    #popt = [0.4132, -0.7011, 0.02891, -0.1637, 1.009] #<20190207># fixed xCOLDGASS Z_PP04_O3N2
#    #popt = [0.4496, -0.6859, 0.02930, -0.1602, 0.9698] #<20190208># still fitting Bertemes and Lee's data
#    popt = [0.4635, -0.7033, 0.03099, -0.1462, 0.8439] #<20190301># Eq850 this work, RemyRuyer GDR, Genzel2015_Eq12a_with_dzliu_limit MZR, KMT09 fmol, fitting all available data '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plots/Plot_z_deltaGas_data_fitting/20190301_with_gas_mass_calibration_Eq850_dzliu/20190301_fitting_dzliu_function/'
#    a, b, c, d, e = popt
#    log10_Ratio_M_molgas_M_star = a*DeltaMS + (b+c*cosmoAge)*(lgMstar-10.0) + d*(cosmoAge) + e
#    Ratio_M_molgas_M_star = 10**log10_Ratio_M_molgas_M_star
#    return Ratio_M_molgas_M_star

#def calc_gas_fraction_A3COSMOS_with_dzliu_850_gas_mass_calibration(cosmoAge, lgMstar=10.5, DeltaMS=0.0):
#    # A3COSMOS 
#    # Sample 20181203 
#    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plots/Plot_z_deltaGas_data_fitting/20190122_with_Leroy_GDR_with_Genzel_Eq12a_MZR_with_dzliu_gas_mass_calibration_with_KMT09_fmol/best_fit_via_CurveFit.json'
#    #popt = [0.7176359685433024, -0.3304845603259402, 0.008136603197619791, -0.07859177961327789, -0.026455117351337622] #<20190122_with_Leroy_GDR_with_Genzel_Eq12a_MZR_with_KMT09_fmol>#
#    popt = [0.4635, -0.7033, 0.03099, -0.1462, 0.8439] #<20190301># Eq850 this work, RemyRuyer GDR, Genzel2015_Eq12a_with_dzliu_limit MZR, KMT09 fmol, fitting all available data '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plots/Plot_z_deltaGas_data_fitting/20190301_with_gas_mass_calibration_Eq850_dzliu/20190301_fitting_dzliu_function/'
#    a, b, c, d, e = popt
#    log10_Ratio_M_molgas_M_star = a*DeltaMS + (b+c*cosmoAge)*(lgMstar-10.0) + d*(cosmoAge) + e
#    Ratio_M_molgas_M_star = 10**log10_Ratio_M_molgas_M_star
#    return Ratio_M_molgas_M_star






# 
# def core functions (adapted from "a_dzliu_code_fit_z_tauDepl_v6.py")
# 
def func_deltaGas_dzliu_log(pars, a, b, c, d, e): 
    if type(pars) is dict:
        for t in ['tauDepl','DeltaMS','lgMstar','lgSFR','cosmoAge','z']:
            if t in pars:
                locals()[t] = copy.copy(pars[t])
    else:
        tauDepl,DeltaMS,lgMstar,lgSFR,cosmoAge,z = pars
    # 20180331a
    # deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*numpy.power(numpy.log10(1+z)) + d
    # 20180331b
    # deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*numpy.power((1+z),d) + e # NOT GOOD
    # 20180331c
    # deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*numpy.power(numpy.log10(1+z),1) + d*numpy.power(numpy.log10(1+z),2) + e
    # 20180331d
    # deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*numpy.power(numpy.log10(1+z),d) + e
    # 20180331e
    # deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0)
    # deltaGas_model = numpy.power(10, deltaGas_model_log) * e * numpy.power(z/c, d) * numpy.exp(-z/c)
    # 20180331f
    #deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0)
    #deltaGas_model = numpy.power(10, deltaGas_model_log) * e * (z - numpy.power(z/c,d))
    # 20180331g
    #cosmoAge = cosmo.age(z).value
    #deltaGas_model_log = a*DeltaMS*(1-numpy.exp(-DeltaMS/d)) + b*(lgMstar-10.0) + c*(cosmoAge) + e
    #deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*(cosmoAge) + d
    #deltaGas_model = numpy.power(10, deltaGas_model_log) # * numpy.power((1+z),c)
    # 20180401a
    #deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*numpy.log(cosmoAge) + d
    #deltaGas_model = numpy.power(10, deltaGas_model_log) # * numpy.power((1+z),c)
    # 20180401b
    #deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*(cosmoAge) + d
    # 20181030a
    #deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*(1-np.exp(cosmoAge)) + d
    # 20181102a
    #deltaGas_model_log = (np.log10(tauDepl))*0.0 + a*DeltaMS + b*(lgSFR-2.0)*0.0 + c*(lgMstar-10.0) + d*(cosmoAge) + e
    deltaGas_model_log = (np.log10(tauDepl))*0.0 + a*DeltaMS + (lgSFR-2.0)*0.0 + (b+c*cosmoAge)*(lgMstar-10.0) + d*(cosmoAge) + e
    # 
    return deltaGas_model_log

def func_deltaGas_dzliu(pars, a, b, c, d, e): 
    if type(pars) is dict:
        deltaGas_model_log = func_deltaGas_dzliu_log(pars, a, b, c, d, e)
    else:
        tauDepl,DeltaMS,lgMstar,lgSFR,cosmoAge,z = pars
        deltaGas_model_log = func_deltaGas_dzliu_log((tauDepl,DeltaMS,lgMstar,lgSFR,cosmoAge,z), a, b, c, d, e)
    deltaGas_model = numpy.power(10, deltaGas_model_log)
    return deltaGas_model

def func_deltaGas_Tacconi2018_log(pars, a, b, c, d): 
    DeltaMS,lgMstar,z = pars
    deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.7) + c*(numpy.log10(1+z)-0.66)**2 + d
    return deltaGas_model_log

def func_deltaGas_Tacconi2018(pars):
    DeltaMS,lgMstar,z = pars
    deltaGas_model_log = func_deltaGas_Tacconi2018_log((DeltaMS,lgMstar,z), 0.53, -0.35, -3.62, 0.12)
    deltaGas_model = numpy.power(10, deltaGas_model_log) # * numpy.power((1+z),c)
    return deltaGas_model

def func_deltaGas_Scoville2017_log(pars, a, b, c, d): 
    DeltaMS,lgMstar,z = pars
    deltaGas_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*(numpy.log10(1+z)) + d
    return deltaGas_model_log

def func_deltaGas_Scoville2017(pars):
    DeltaMS,lgMstar,z = pars
    deltaGas_model_log = func_deltaGas_Scoville2017_log((DeltaMS,lgMstar,z), 0.32, -0.70, 1.84, numpy.log10(0.71))
    deltaGas_model = numpy.power(10, deltaGas_model_log) # * numpy.power((1+z),c)
    return deltaGas_model











#!/usr/bin/env python
# 
# 20190222
#     copied from "calc_stellar_mass_function.py", this code will superceed "calc_stellar_mass_function.py". 
# 

from __future__ import print_function

import os, sys, re, json, time, astropy
import numpy as np
from astropy.table import Table, Column, hstack
from copy import copy
from numpy import log, log10, power, sum, sqrt, pi, exp
pow = power
lg = log10
ln = log
from scipy.interpolate import InterpolatedUnivariateSpline, interp1d

if not (os.path.dirname(os.path.abspath(__file__)) in sys.path): sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import apply_cosmology
cosmo = apply_cosmology.cosmo

if sys.version_info.major >= 3:
    long = int
else:
    pass




# 
# def 
# 
def Schechter_Function(lgM, phi, lg_M0, alpha):
    # 
    # Schechter (1976)
    # 
    # Phi(M) dM = (Phi_*) * (M/M_*)**(alpha) * exp(-M/M_*) dM/M_*
    #           = (Phi_*) * x**(alpha) * exp(-x) dx
    #           = (Phi_*) * 10**(lgx * alpha) * exp(-10**lgx) dx
    #           = (Phi_*) * 10**(lgx * alpha) * exp(-10**lgx) dlnx
    #           = (Phi_*) * 10**(lgx * alpha) * exp(-10**lgx) dlgx * ln(10)
    #           = (Phi_*) * 10**((lgM-lgM_*)*(alpha+1)) * exp(-10**(lgM-lgM_*)) * ln(10)  dlgx
    #           = (Our_Phi_Phi_Schechter)                                                 dlgx
    # 
    lgx = lgM-lg_M0
    Phi_Schechter = phi * (10**(lgx*(alpha+1))) * (np.exp(-10**lgx)) * ln(10) # per dex and already multiplied ln(10), so that its integral directly equals \int Phi(M) / M dM
    return Phi_Schechter




# 
# def 
# 
def calc_SMF_Davidzon2017(z, lgMstar=None, galaxy_type = 'SFG'):
    # 
    # Davidzon 2017 arXiv:1701.02734
    # IMF: Chabrier 2003
    # Outputs: lgMstar_grid, lgPhiMstar_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['ALL', 'SFG', 'QG']):
            raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    # 
    # make lgMstar
    if lgMstar is None:
        lgMstar_grid = np.linspace(6.0, 13.0, num=1000, endpoint=True)
    else:
        lgMstar_grid = lgMstar
    # 
    # read SMF
    tb_SMF = Table.read(os.path.dirname(os.path.dirname(__file__))+os.sep+'Data_Tables/datatables_SMF/datatable_Davidzon2017_SMF_'+galaxy_type+'.txt', format='ascii') # zLo zHi lgMchar  Phi_1   alpha_1  Phi_2  alpha_2
    SMF_zmin = np.min(tb_SMF['zLo'])
    SMF_zmax = np.max(tb_SMF['zHi'])
    # 
    # check z
    if z < SMF_zmin or z > SMF_zmax:
        raise ValueError('calc_SMF_Davidzon2017: The input redshift is out of the allowed range of %s -- %s!'%(SMF_zmin, SMF_zmax))
    # 
    # spline SMF #<20190214># old method
    #<20190214># SMF_z = (tb_SMF['zLo'].data + tb_SMF['zHi'].data) / 2.0
    #<20190214># SMF_phi_1 = InterpolatedUnivariateSpline(SMF_z, tb_SMF['Phi_1'].data, k=1)(z)
    #<20190214># SMF_phi_2 = InterpolatedUnivariateSpline(SMF_z, tb_SMF['Phi_2'].data, k=1)(z)
    #<20190214># SMF_alpha_1 = InterpolatedUnivariateSpline(SMF_z, tb_SMF['alpha_1'].data, k=1)(z)
    #<20190214># SMF_alpha_2 = InterpolatedUnivariateSpline(SMF_z, tb_SMF['alpha_2'].data, k=1)(z)
    #<20190214># SMF_lgMchar = InterpolatedUnivariateSpline(SMF_z, tb_SMF['lgMchar'].data, k=1)(z)
    #<20190214># #print('z, lgMchar, alpha_1, phi_1, alpha_2, phi_2 =', z, SMF_lgMchar, SMF_alpha_1, SMF_phi_1, SMF_alpha_2, SMF_phi_2)
    #<20190214># SMF_PhiMstar = Schechter_Function(lgMstar_grid, SMF_phi_1, SMF_lgMchar, SMF_alpha_1) + \
    #<20190214>#                Schechter_Function(lgMstar_grid, SMF_phi_2, SMF_lgMchar, SMF_alpha_2) # two component
    #<20190214># lgPhiMstar_grid = np.log10(SMF_PhiMstar)
    # 
    # spline SMF
    lgPhiMstar_matrix = []
    for k in range(len(tb_SMF)):
        SMF_z = (tb_SMF['zLo'][k] + tb_SMF['zHi'][k]) / 2.0
        SMF_phi_1 = tb_SMF['Phi_1'][k]
        SMF_phi_2 = tb_SMF['Phi_2'][k]
        SMF_alpha_1 = tb_SMF['alpha_1'][k]
        SMF_alpha_2 = tb_SMF['alpha_2'][k]
        SMF_lgMchar = tb_SMF['lgMchar'][k]
        #print('z, lgMchar, alpha_1, phi_1, alpha_2, phi_2 =', z, SMF_lgMchar, SMF_alpha_1, SMF_phi_1, SMF_alpha_2, SMF_phi_2)
        SMF_PhiMstar = Schechter_Function(lgMstar_grid, SMF_phi_1, SMF_lgMchar, SMF_alpha_1) + \
                       Schechter_Function(lgMstar_grid, SMF_phi_2, SMF_lgMchar, SMF_alpha_2) # two component
        lgPhiMstar_grid = np.log10(SMF_PhiMstar)
        lgPhiMstar_matrix.append(copy(lgPhiMstar_grid))
    # 
    SMF_z = (tb_SMF['zLo'].data + tb_SMF['zHi'].data) / 2.0
    lgPhiMstar_matrix = np.array(lgPhiMstar_matrix) # shape == (N_SMF_z, N_SMF_lgMstar, )
    if z <= np.min(SMF_z):
        lgPhiMstar_grid = lgPhiMstar_matrix[0]
    elif z >= np.max(SMF_z):
        lgPhiMstar_grid = lgPhiMstar_matrix[-1]
    else:
        lgPhiMstar_grid = interp1d(SMF_z, lgPhiMstar_matrix, axis=0, kind='linear')(z)
    #print(lgPhiMstar_matrix.shape, SMF_z.shape, lgPhiMstar_grid.shape)
    # fix nan
    lgPhiMstar_grid[np.isnan(lgPhiMstar_grid)] = -100
    lgPhiMstar_grid[(lgPhiMstar_grid<-100)] = -100
    # 
    if lgMstar is None:
        return lgMstar_grid, lgPhiMstar_grid
    else:
        return lgPhiMstar_grid



def calc_SMF_Moutard2016(z, lgMstar=None, galaxy_type = 'SFG'):
    # 
    # Moutard 2016 - SMF - https://ui.adsabs.harvard.edu/abs/2016A%26A...590A.103M/abstract
    # IMF: Chabrier 2003
    # Outputs: lgMstar_grid, lgPhiMstar_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['ALL', 'SFG', 'QG']):
            raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    # 
    # make lgMstar
    if lgMstar is None:
        lgMstar_grid = np.linspace(6.0, 13.0, num=1000, endpoint=True)
    else:
        lgMstar_grid = lgMstar
    # 
    # read SMF
    tb_SMF = Table.read(os.path.dirname(os.path.dirname(__file__))+os.sep+'Data_Tables/datatables_SMF/datatable_Moutard2016_SMF_'+galaxy_type+'.txt', format='ascii') # zLo zHi lgMchar  Phi_1   alpha_1  Phi_2  alpha_2
    SMF_zmin = np.min(tb_SMF['zLo'])
    SMF_zmax = np.max(tb_SMF['zHi'])
    # 
    # check z
    if z < SMF_zmin or z > SMF_zmax:
        raise ValueError('calc_SMF_Moutard2016: The input redshift is out of the allowed range of %s -- %s!'%(SMF_zmin, SMF_zmax))
    # 
    # spline SMF
    lgPhiMstar_matrix = []
    for k in range(len(tb_SMF)):
        SMF_z = (tb_SMF['zLo'][k] + tb_SMF['zHi'][k]) / 2.0
        SMF_phi_1 = tb_SMF['Phi_1'][k]
        SMF_phi_2 = tb_SMF['Phi_2'][k]
        SMF_alpha_1 = tb_SMF['alpha_1'][k]
        SMF_alpha_2 = tb_SMF['alpha_2'][k]
        SMF_lgMchar = tb_SMF['lgMchar'][k]
        #print('calc_SMF_Moutard2016: z %r, lgMchar %r, alpha_1 %r, phi_1 %r, alpha_2 %r, phi_2 %r'%(z, SMF_lgMchar, SMF_alpha_1, SMF_phi_1, SMF_alpha_2, SMF_phi_2))
        SMF_PhiMstar = Schechter_Function(lgMstar_grid, SMF_phi_1, SMF_lgMchar, SMF_alpha_1) + \
                       Schechter_Function(lgMstar_grid, SMF_phi_2, SMF_lgMchar, SMF_alpha_2) # two component
        lgPhiMstar_grid = np.log10(SMF_PhiMstar)
        lgPhiMstar_matrix.append(copy(lgPhiMstar_grid))
    # 
    SMF_z = (tb_SMF['zLo'].data + tb_SMF['zHi'].data) / 2.0
    lgPhiMstar_matrix = np.array(lgPhiMstar_matrix) # shape == (N_SMF_z, N_SMF_lgMstar, )
    if z <= np.min(SMF_z):
        lgPhiMstar_grid = lgPhiMstar_matrix[0]
    elif z >= np.max(SMF_z):
        lgPhiMstar_grid = lgPhiMstar_matrix[-1]
    else:
        lgPhiMstar_grid = interp1d(SMF_z, lgPhiMstar_matrix, axis=0, kind='linear')(z)
    #print(lgPhiMstar_matrix.shape, SMF_z.shape, lgPhiMstar_grid.shape)
    # fix nan
    lgPhiMstar_grid[np.isnan(lgPhiMstar_grid)] = -100
    lgPhiMstar_grid[(lgPhiMstar_grid<-100)] = -100
    # 
    if lgMstar is None:
        return lgMstar_grid, lgPhiMstar_grid
    else:
        return lgPhiMstar_grid



def calc_SMF_Ilbert2013(z, lgMstar=None, galaxy_type = 'SFG'):
    # 
    # Ilbert 2013
    # IMF: Chabrier 2003
    # Outputs: lgMstar_grid, lgPhiMstar_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL" or "SFG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['ALL', 'SFG', 'QG']):
            raise ValueError('Please input either "ALL" or "SFG" as the galaxy_type!')
    # 
    # make lgMstar
    if lgMstar is None:
        lgMstar_grid = np.linspace(6.0, 13.0, num=1000, endpoint=True)
    else:
        lgMstar_grid = lgMstar
    # 
    # read SMF
    tb_SMF = Table.read(os.path.dirname(os.path.dirname(__file__))+os.sep+'Data_Tables/datatables_SMF/datatable_Ilbert2013_SMF_'+galaxy_type+'.txt', format='ascii') # zLo zHi lgMchar  Phi_1   alpha_1  Phi_2  alpha_2
    SMF_zmin = np.min(tb_SMF['zLo'])
    SMF_zmax = np.max(tb_SMF['zHi'])
    # 
    # check z
    if z < SMF_zmin or z > SMF_zmax:
        raise ValueError('calc_SMF_Ilbert2013: The input redshift is out of the allowed range of %s -- %s!'%(SMF_zmin, SMF_zmax))
    # 
    # spline SMF
    lgPhiMstar_matrix = []
    for k in range(len(tb_SMF)):
        SMF_z = (tb_SMF['zLo'][k] + tb_SMF['zHi'][k]) / 2.0
        SMF_phi_1 = tb_SMF['Phi_1'][k]
        SMF_phi_2 = tb_SMF['Phi_2'][k]
        SMF_alpha_1 = tb_SMF['alpha_1'][k]
        SMF_alpha_2 = tb_SMF['alpha_2'][k]
        SMF_lgMchar = tb_SMF['lgMchar'][k]
        #print('z, lgMchar, alpha_1, phi_1, alpha_2, phi_2 =', z, SMF_lgMchar, SMF_alpha_1, SMF_phi_1, SMF_alpha_2, SMF_phi_2)
        SMF_PhiMstar = Schechter_Function(lgMstar_grid, SMF_phi_1, SMF_lgMchar, SMF_alpha_1) + \
                       Schechter_Function(lgMstar_grid, SMF_phi_2, SMF_lgMchar, SMF_alpha_2) # two component
        lgPhiMstar_grid = np.log10(SMF_PhiMstar)
        lgPhiMstar_matrix.append(copy(lgPhiMstar_grid))
    # 
    SMF_z = (tb_SMF['zLo'].data + tb_SMF['zHi'].data) / 2.0
    lgPhiMstar_matrix = np.array(lgPhiMstar_matrix) # shape == (N_SMF_z, N_SMF_lgMstar, )
    if z <= np.min(SMF_z):
        lgPhiMstar_grid = lgPhiMstar_matrix[0]
    elif z >= np.max(SMF_z):
        lgPhiMstar_grid = lgPhiMstar_matrix[-1]
    else:
        lgPhiMstar_grid = interp1d(SMF_z, lgPhiMstar_matrix, axis=0, kind='linear')(z)
    #print(lgPhiMstar_matrix.shape, SMF_z.shape, lgPhiMstar_grid.shape)
    # fix nan
    lgPhiMstar_grid[np.isnan(lgPhiMstar_grid)] = -100
    lgPhiMstar_grid[(lgPhiMstar_grid<-100)] = -100
    # 
    if lgMstar is None:
        return lgMstar_grid, lgPhiMstar_grid
    else:
        return lgPhiMstar_grid



def calc_SMF_Peng2010(z, lgMstar=None, galaxy_type='SFG'):
    # 
    # Peng YingJie 2010
    # IMF: Chabrier 2003
    # Outputs: lgMstar_grid, lgPhiMstar_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['SFG', 'QG']):
            raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    # 
    # make lgMstar
    if lgMstar is None:
        lgMstar_grid = np.linspace(6.0, 13.0, num=1000, endpoint=True)
    else:
        lgMstar_grid = lgMstar
    # 
    # galaxy_type
    if galaxy_type == 'ALL':
        galaxy_types = ['SFG', 'QG']
    else:
        galaxy_types = [galaxy_type]
    # 
    # read SMF
    for t_galaxy_type in galaxy_types:
        tb_SMF = Table.read(os.path.dirname(os.path.dirname(__file__))+os.sep+'Data_Tables/datatables_SMF/datatable_PengYingjie2010_SMF_'+galaxy_type+'.txt', format='ascii') # zLo zHi lgMchar  Phi_1   alpha_1  Phi_2  alpha_2
        SMF_zmin = np.min(tb_SMF['zLo'])
        SMF_zmax = np.max(tb_SMF['zHi'])
        # 
        # there is only one redshift bin, but we still check the input z range
        if z < SMF_zmin or z > SMF_zmax:
            raise ValueError('calc_SMF_Peng2010: The input redshift is out of the allowed range of %s -- %s!'%(SMF_zmin, SMF_zmax))
        # 
        # just calculate SMF without interpolation
        SMF_z = (tb_SMF['zLo'].data + tb_SMF['zHi'].data) / 2.0
        SMF_phi_1 = tb_SMF['Phi_1'].data
        SMF_alpha_1 = tb_SMF['alpha_1'].data
        SMF_lgMchar = tb_SMF['lgMchar'].data
        SMF_PhiMstar = Schechter_Function(lgMstar_grid, SMF_phi_1, SMF_lgMchar, SMF_alpha_1) # one component
        if t_galaxy_type == 'SFG':
            SMF_PhiMstar_SFG = copy(SMF_PhiMstar)
        elif t_galaxy_type == 'QG':
            SMF_phi_2 = tb_SMF['Phi_2'].data
            SMF_alpha_2 = tb_SMF['alpha_2'].data
            SMF_PhiMstar_QG = SMF_PhiMstar + Schechter_Function(lgMstar_grid, SMF_phi_2, SMF_lgMchar, SMF_alpha_2) # two component QG SMF
    # 
    if galaxy_type == 'SFG':
        lgPhiMstar_grid = np.log10(SMF_PhiMstar_SFG)
    elif galaxy_type == 'QG':
        lgPhiMstar_grid = np.log10(SMF_PhiMstar_QG)
    elif galaxy_type == 'ALL':
        lgPhiMstar_grid = np.log10(SMF_PhiMstar_SFG+SMF_PhiMstar_QG)
    # 
    if lgMstar is None:
        return lgMstar_grid, lgPhiMstar_grid
    else:
        return lgPhiMstar_grid



def calc_SMF_Kelvin2014(z, lgMstar=None, galaxy_type='SFG'):
    # 
    # Kelvin 2014 (2014MNRAS.444.1647K)
    # GAMA survey
    # IMF: Chabrier 2003
    # Table 3, Spheroid dominated, and Disc dominated
    # Outputs: lgMstar_grid, lgPhiMstar_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['SFG', 'QG']):
            raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    # 
    # make lgMstar
    if lgMstar is None:
        lgMstar_grid = np.linspace(6.0, 13.0, num=1000, endpoint=True)
    else:
        lgMstar_grid = lgMstar
    # 
    # galaxy_type
    if galaxy_type == 'ALL':
        galaxy_types = ['SFG', 'QG']
    else:
        galaxy_types = [galaxy_type]
    # 
    # read SMF
    for t_galaxy_type in galaxy_types:
        tb_SMF = Table.read(os.path.dirname(os.path.dirname(__file__))+os.sep+'Data_Tables/datatables_SMF/datatable_Kelvin2014_SMF_'+galaxy_type+'.txt', format='ascii') # zLo zHi lgMchar Phi_1 alpha_1
        SMF_zmin = np.min(tb_SMF['zLo'])
        SMF_zmax = np.max(tb_SMF['zHi'])
        # 
        # there is only one redshift bin, but we still check the input z range
        if z < SMF_zmin or z > SMF_zmax:
            raise ValueError('calc_SMF_Kelvin2014: The input redshift is out of the allowed range of %s -- %s!'%(SMF_zmin, SMF_zmax))
        # 
        # just calculate SMF without interpolation
        SMF_z = (tb_SMF['zLo'].data + tb_SMF['zHi'].data) / 2.0
        SMF_phi_1 = tb_SMF['Phi_1'].data
        SMF_alpha_1 = tb_SMF['alpha_1'].data
        SMF_lgMchar = tb_SMF['lgMchar'].data
        SMF_PhiMstar = Schechter_Function(lgMstar_grid, SMF_phi_1, SMF_lgMchar, SMF_alpha_1) # one component
        if t_galaxy_type == 'SFG':
            SMF_PhiMstar_SFG = copy(SMF_PhiMstar)
        elif t_galaxy_type == 'QG':
            SMF_PhiMstar_QG = copy(SMF_PhiMstar)
    # 
    if galaxy_type == 'SFG':
        lgPhiMstar_grid = np.log10(SMF_PhiMstar_SFG)
    elif galaxy_type == 'QG':
        lgPhiMstar_grid = np.log10(SMF_PhiMstar_QG)
    elif galaxy_type == 'ALL':
        lgPhiMstar_grid = np.log10(SMF_PhiMstar_SFG+SMF_PhiMstar_QG)
    # 
    if lgMstar is None:
        return lgMstar_grid, lgPhiMstar_grid
    else:
        return lgPhiMstar_grid



def calc_SMF_Wright2018_single_component(z, lgMstar=None):
    # 
    # Wright 2018 - SMF - arXiv.1808.01754 - 2018MNRAS.480.3491W.pdf
    # IMF: Chabrier 2003
    # Outputs: lgMstar_grid, lgPhiMstar_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # make lgMstar
    if lgMstar is None:
        lgMstar_grid = np.linspace(6.0, 13.0, num=1000, endpoint=True)
    else:
        lgMstar_grid = lgMstar
    # 
    # 
    SMF_lgMchar = (10.791) + (0.558) * z + (-0.102) * z**2
    SMF_alpha = (-1.160) + (-0.274) * z + (0.028) * z**2
    SMF_phi = 10**((-2.455) + (-0.883) * z + (0.093) * z**2)
    #SMF_rho = (-2.455) + (-0.883) * z + (0.093) * z**2
    #print('z, lgMchar, alpha, phi =', z, SMF_lgMchar, SMF_alpha, SMF_phi)
    SMF_PhiMstar = Schechter_Function(lgMstar_grid, SMF_phi, SMF_lgMchar, SMF_alpha)
    lgPhiMstar_grid = np.log10(SMF_PhiMstar)
    # 
    if lgMstar is None:
        return lgMstar_grid, lgPhiMstar_grid
    else:
        return lgPhiMstar_grid


def calc_SMF_Wright2018_double_component(z, lgMstar=None):
    # 
    # Wright 2018 - SMF - arXiv.1808.01754 - 2018MNRAS.480.3491W.pdf
    # IMF: Chabrier 2003
    # Outputs: lgMstar_grid, lgPhiMstar_grid
    # 
    # check z
    if not np.isscalar(z):
        raise ValueError('Please input a float number as the redshift!')
    # 
    # make lgMstar
    if lgMstar is None:
        lgMstar_grid = np.linspace(6.0, 13.0, num=1000, endpoint=True)
    else:
        lgMstar_grid = lgMstar
    # 
    # 
    SMF_lgMchar = (10.831) + (0.153) * z + (-0.033) * z**2
    SMF_alpha_1 = (-0.579) + (0.048) * z + (0.022) * z**2
    SMF_alpha_2 = (-1.489) + (-0.087) * z + (0.016) * z**2
    SMF_phi_1 = 10**((-2.312) + (-0.658) * z + (0.016) * z**2)
    SMF_phi_2 = 10**((-3.326) + (-0.158) * z + (-0.002) * z**2)
    #print('z, lgMchar, alpha_1, phi_1, alpha_2, phi_2 =', z, SMF_lgMchar, SMF_alpha_1, SMF_phi_1, SMF_alpha_2, SMF_phi_2)
    SMF_PhiMstar = Schechter_Function(lgMstar_grid, SMF_phi_1, SMF_lgMchar, SMF_alpha_1) + \
                   Schechter_Function(lgMstar_grid, SMF_phi_2, SMF_lgMchar, SMF_alpha_2)
    lgPhiMstar_grid = np.log10(SMF_PhiMstar)
    # 
    if lgMstar is None:
        return lgMstar_grid, lgPhiMstar_grid
    else:
        return lgPhiMstar_grid










def calc_Gladders2013_CSFRD(z, t0, tau):
    Gladders2013_t_age = cosmo.age(z).value
    Gladders2013_t0 = float(t0)
    Gladders2013_tau = float(tau)
    Gladders2013_SFR_1 = 1.0/(Gladders2013_t_age*sqrt(2*pi*Gladders2013_tau**2))
    Gladders2013_SFR_2 = exp(-(ln(Gladders2013_t_age)-Gladders2013_t0)**2/(2*Gladders2013_tau**2))
    Gladders2013_SFR = Gladders2013_SFR_1 * Gladders2013_SFR_2
    Gladders2013_SFR = Gladders2013_SFR / 1.64 # converted to Chabrier IMF
    return Gladders2013_SFR


def calc_MadauDickinson2014_CSFRD(z):
    # Madau & Dickinson (2014)
    # Salpeter IMF --> converted to Chabrier IMF
    if type(z) is list:
        z = np.array(z)
    rho_SFR = 0.015 * (1+z)**2.7 / (1.0 + ((1+z)/2.9)**5.6) / 1.64 # converted to Chabrier IMF
    return rho_SFR


def calc_Mstar_integrating_CSFRD_dzliu2018(z):
    # make z_list
    #z_list = np.arange(9.75, 0.75, -0.5).tolist()
    #z_list.extend([0.75, 0.50, 0.25, 0.125, 0.0625, 0.03125, 0.00000])
    opz_list = np.logspace(np.log10(1.0+0.0), np.log10(1.0+10.75), num=200, endpoint=True)
    opz_list = opz_list[::-1]
    z_list = opz_list - 1.0
    t_list = cosmo.age(z_list).value
    # 
    # integrate dzliu2018 CSFRD
    #CSFRD = calc_Gladders2013_CSFRD(z_list, 1.57, 0.67) # Gladder2013 form of CSFRD with Liu et al. (2018) parameters
    CSFRD = calc_MadauDickinson2014_CSFRD(z_list)
    Mstar_cumulated = 0.0
    CSFRD_z_list = [z_list[0]]
    CSFRD_Mstar_list = [1e-30] #<TODO># initial mass
    for i in range(len(z_list)-1):
        #z_bin = (z_list[i+1] + z_list[i]) / 2.0
        t_bin = t_list[i+1] # Gyr, time since the Big Bang
        time_bin = t_list[i+1] - t_list[i] # Gyr, time interval within the redshift bin
        #mass_loss_time_scale = 1.4 # Myr, BEHROOZI & Silk arXiv:1404.5299
        #mass_loss_time_scale = 3.0 # Myr, Ilbert et al. (2013) PDF page 11 left middle
        mass_loss_time_scale = 0.3 # Myr, Conroy & Wechsler (2009, bibcode 2009ApJ...696..620C) arxiv PDF page 5 Eq (11). 
        Mstar_formed = (CSFRD[i]+CSFRD[i+1])/2.0 * time_bin * 1e9 # Msun/yr * Gyr * 1e9 => Msun
        Mstar_loss_frac = 0.05 * ln(1.0+(t_bin)/(mass_loss_time_scale*1e-3)) 
                                # see Ilbert et al. (2013) PDF page 11 left middle; Conroy & Wechsler (2009) arxiv PDF page 5 Eq (11). 
                                # see https://arxiv.org/pdf/1404.5299.pdf PDF page 3 Eq (6); Conroy & Wechsler (2009) arxiv PDF page 5 Eq (11). 
        #Mstar_loss_frac = 0.0 # no mass loss at all <TODO>
        Mstar_cumulated += Mstar_formed * (1.0 - Mstar_loss_frac) # total new star created, locked in stellar mass
        #CSFRD_z_list.append(z_bin)
        CSFRD_z_list.append(z_list[i+1])
        CSFRD_Mstar_list.append(Mstar_cumulated)
    CSFRD_z_list = np.array(CSFRD_z_list)[::-1] # sort so that z increases
    CSFRD_Mstar_list = np.array(CSFRD_Mstar_list)[::-1]
    # spline at z for the output
    Mstar_cumulated_at_z = 10**(InterpolatedUnivariateSpline(CSFRD_z_list, np.log10(CSFRD_Mstar_list), k=1)(z))
    return Mstar_cumulated_at_z



def calc_SMF_dzliu2018(z=None, lgMstar=None, galaxy_type='SFG', z_list=None, tuning_params='', verbose=True):
    # 
    # dzliu 2018 - my own choice
    # IMF: Chabrier 2003
    # Outputs: lgMstar_grid, lgPhiMstar_grid
    # 
    # tuning_params: for example, 'D17-no-renorm;'
    # 
    # 
    # check z
    if z is not None:
        if not np.isscalar(z):
            #raise ValueError('Please input a float number as the redshift!')
            if type(z) is list:
                z = np.array(z)
            z_is_vector = True
        else:
            z_is_vector = False
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['SFG', 'QG', 'ALL']):
            raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    # 
    # make lgMstar
    if lgMstar is None:
        lgMstar_grid = np.linspace(6.0, 13.0, num=1000, endpoint=True)
    else:
        lgMstar_grid = lgMstar
    # 
    # code from '/Volumes/GoogleDrive/Team Drives/DeepFields/Simulations/Cosmological_Galaxy_Modelling/a_dzliu_code_Plot_SMF_dzliu_model.sm'
    # make z_list
    if z_list is None:
        ##z_list = np.arange(10.75, 0.75, -0.5).tolist()
        ##z_list.extend([0.75, 0.50, 0.25, 0.125, 0.0625, 0.03125, 0.015625, 0.00000])
        #z_list = np.arange(10.75, 1.00, -0.5).tolist()
        #z_list.extend([1.00, 0.75, 0.50, 0.25, 0.125, 0.0625, 0.03125, 0.015625, 0.00000])
        #print('z_list', z_list, '(len=%d)'%(len(z_list)))
        # 
        # <20200105> precision issue
        # <20200105> now we increase the sampling of z_list from 29 to 40, so that the SMF are better interpolated.
        z_list = [] 
        z_list.extend(np.arange(10.75, 5.00, -0.5).tolist())
        z_list.extend(np.arange(5.00, 1.00, -0.25).tolist())
        z_list.extend(np.arange(1.00, 0.125, -0.125).tolist())
        z_list.extend([0.125, 0.0625, 0.03125, 0.015625, 0.00000])
        #print('z_list', z_list, '(len=%d)'%(len(z_list)))
        # <20200105> precision issue (end)
        # 
        #z_list = []
        #z_list.extend(np.arange(9.75, 4.00, -0.5).tolist())
        #z_list.extend(np.arange(4.00, 3.00, -0.25).tolist())
        #z_list.extend(np.arange(3.00, 2.00, -0.125).tolist())
        #z_list.extend(np.arange(2.00, 1.00, -0.0625).tolist())
        #z_list.extend(np.arange(1.00, 0.00, -0.03125).tolist())
        #z_list.extend([0.00])
    SMF_z_list = []
    SMF_Phi_list = []
    Mstar_cumulated_list = calc_Mstar_integrating_CSFRD_dzliu2018(z_list)
    for i in range(len(z_list)-1):
        z_bin = (z_list[i] + z_list[i+1]) / 2.0
        t_bin = cosmo.age(z_bin).value # Gyr since the Big Bang
        Schechter_M = 10**lgMstar_grid
        Schechter_Mstep = lgMstar_grid[1] - lgMstar_grid[0]
        do_renorm_by_CSFRD_cumulated_Mstar = True
        # 
        # first construct SMF at each redshift bin
        #if False:
        #    EEE_Mchar_MQG = lg((10**10.68)*(t_bin/cosmo.age(0).value)**(0.50)) # Quiescent galaxies' SMF's first component
        #    EEE_Mchar_NQG = lg((10**10.68)*(t_bin/cosmo.age(0).value)**(0.50)) # Quiescent galaxies' SMF's second component
        #    EEE_Mchar_SFG = lg((10**10.72)*(t_bin/cosmo.age(0).value)**(-0.50)) # Star-forming galaxies' SMF
        #    EEE_Phi_MQG = (3.400e-3)*(t_bin/cosmo.age(0).value)**(4.00)
        #    EEE_Phi_NQG = (0.126e-3)*(t_bin/cosmo.age(0).value)**(4.00) # decrease much faster with increasing z
        #    EEE_Phi_SFG = (0.900e-3)*(t_bin/cosmo.age(0).value)**(0.20)
        #    EEE_alpha_MQG = -0.39 + (z_bin)*(1.00)
        #    EEE_alpha_NQG = -1.56 + (z_bin)*(1.00)
        #    EEE_alpha_SFG = -1.40 + (z_bin)*(-0.06) # slope -- this makes too many low mass galaxiest at z>4 -- faint-end of the 24um number counts -- no, this is because stellar SED has too old age, f24um is underestimated!
        #    #EEE_alpha_SFG = -1.40 + (z)*(-0.03) # slope
        #    Schechter_P_MQG = Schechter_Function(lgMstar_grid, EEE_Phi_MQG, EEE_Mchar_MQG, EEE_alpha_MQG)
        #    Schechter_P_NQG = Schechter_Function(lgMstar_grid, EEE_Phi_NQG, EEE_Mchar_NQG, EEE_alpha_NQG)
        #    Schechter_P_SFG = Schechter_Function(lgMstar_grid, EEE_Phi_SFG, EEE_Mchar_SFG, EEE_alpha_SFG)
        #    Schechter_P_QG = Schechter_P_MQG + Schechter_P_NQG
        #    #print('z = %.04f, lgSchechter_P_SFG = %s, lgSchechter_P_QG = %s'%(z_bin, np.log10(Schechter_P_SFG), np.log10(Schechter_P_QG)))
        # 
        if z_bin < 0.02:
            Schechter_P_QG = 10**(calc_SMF_Peng2010(0.02, lgMstar=lgMstar_grid, galaxy_type='QG'))
            Schechter_P_SFG = 10**(calc_SMF_Peng2010(0.02, lgMstar=lgMstar_grid, galaxy_type='SFG'))
        elif z_bin < 0.085:
            Schechter_P_QG = 10**(calc_SMF_Peng2010(z_bin, lgMstar=lgMstar_grid, galaxy_type='QG'))
            Schechter_P_SFG = 10**(calc_SMF_Peng2010(z_bin, lgMstar=lgMstar_grid, galaxy_type='SFG'))
        elif z_bin < 0.2:
            Schechter_P_QG = 10**(calc_SMF_Peng2010(0.085, lgMstar=lgMstar_grid, galaxy_type='QG')) * (1.0-((0.2-0.085)-(0.2-z_bin))/(0.2-0.085)) + 10**(calc_SMF_Davidzon2017(0.2, lgMstar=lgMstar_grid, galaxy_type='QG')) * (0.0+((0.2-0.085)-(0.2-z_bin))/(0.2-0.085))
            Schechter_P_SFG = 10**(calc_SMF_Peng2010(0.085, lgMstar=lgMstar_grid, galaxy_type='SFG')) * (1.0-((0.2-0.085)-(0.2-z_bin))/(0.2-0.085)) + 10**(calc_SMF_Davidzon2017(0.2, lgMstar=lgMstar_grid, galaxy_type='SFG')) * (0.0+((0.2-0.085)-(0.2-z_bin))/(0.2-0.085))
        elif z_bin < 4.0:
            Schechter_P_QG = 10**(calc_SMF_Davidzon2017(z_bin, lgMstar=lgMstar_grid, galaxy_type='QG'))
            Schechter_P_SFG = 10**(calc_SMF_Davidzon2017(z_bin, lgMstar=lgMstar_grid, galaxy_type='SFG'))
            #<TODO># QG/SFG fraction seems lower at z~1-3?
            #if z_bin > 1.0 and z_bin < 3.0:
            #    Schechter_P_QG = Schechter_P_QG * np.interp(z_bin, [1.0, 1.5, 2.0, 2.5, 3.0], [1.0, 1.2, 1.2, 1.2, 1.0])
            if tuning_params is not None:
                if tuning_params.find('D17-no-renorm')>=0:
                    do_renorm_by_CSFRD_cumulated_Mstar = False
        else:
            Schechter_P_QG = 10**(calc_SMF_Davidzon2017(4.0, lgMstar=lgMstar_grid, galaxy_type='QG'))
            Schechter_P_SFG = 10**(calc_SMF_Davidzon2017(4.0, lgMstar=lgMstar_grid, galaxy_type='SFG'))
            # 
            # Note that my Schechter_Function already contains * ln(10), i.e., it is per dex!
            # 
        # 
        # 
        # then re-normalize SMF to total stellar mass integrated from CSFRD (assumed some mass loss, see the called function)
        #Mstar_cumulated = calc_Mstar_integrating_CSFRD_dzliu2018(z_bin)
        #Mstar_cumulated = calc_Mstar_integrating_CSFRD_dzliu2018(z_list[i+1]) #<20190915># BUGGY, should be z bin edge, and z_list is in descending order 
        Mstar_cumulated = Mstar_cumulated_list[i+1]
        #CSFRD_at_z_bin = calc_MadauDickinson2014_CSFRD(z_bin)
        #<TODO><20191001># mask = (Schechter_M>=1e9) #<20191001># to match the Madau & Dickinson (2014) integration limit of '0.03 * L_characteristic', i.e., 0.03 * M_characteristic = 0.03 * 10**10.5 = 
        #<TODO><20191001># Schechter_M_total = sum((Schechter_P_QG[mask]+Schechter_P_SFG[mask])*Schechter_M[mask]*Schechter_Mstep) # P is per dex, but our Schechter_Function already contains ln(10), hence we do not need to multiply ln(10) here. \int P(M) dM = \int P(M) M dln(M) = \int P(M)*ln(10) M dlg(M)
        
        if do_renorm_by_CSFRD_cumulated_Mstar:
            Schechter_M_total = sum((Schechter_P_QG+Schechter_P_SFG)*Schechter_M*Schechter_Mstep) # P is per dex, but our Schechter_Function already contains ln(10), hence we do not need to multiply ln(10) here. \int P(M) dM = \int P(M) M dln(M) = \int P(M)*ln(10) M dlg(M)
            renorm_factor = Mstar_cumulated / Schechter_M_total
            Schechter_P_SFG = Schechter_P_SFG * renorm_factor # per dex but already contains ln(10)
            Schechter_P_QG = Schechter_P_QG * renorm_factor # per dex but already contains ln(10)
        Schechter_P_ALL = Schechter_P_SFG + Schechter_P_QG
        if verbose:
            print('z = %.04f, lgMstar_CSFRD = %0.2f, lgMstar_SMF = %0.2f, renorm = %s'%(z_bin, np.log10(Mstar_cumulated), np.log10(Schechter_M_total), renorm_factor))
        #print('z = %.04f, lgCSFRD = %0.2f, lgMstar_CSFRD = %0.2f, lgMstar_SMF = %0.2f'%(z_bin, np.log10(CSFRD_at_z_bin), np.log10(Mstar_cumulated), np.log10(Schechter_M_total)))
        # 
        # --> checked OK
        # z = 0.0156, lgCSFRD = -2.04, lgMstar_CSFRD = 8.40, lgMstar_SMF = 8.35
        # 
        # 
        SMF_z_list.append(z_list[i+1]) # append the lower redshift end point of each bin
        if galaxy_type == 'SFG':
            SMF_Phi_list.append(Schechter_P_SFG)
        elif galaxy_type == 'QG':
            SMF_Phi_list.append(Schechter_P_QG)
        elif galaxy_type == 'ALL':
            SMF_Phi_list.append(Schechter_P_ALL)
        # 
    # spline at z for the output
    SMF_z_list = np.array(SMF_z_list)[::-1] # make z increasing order
    SMF_Phi_list = np.array(SMF_Phi_list)[::-1].T # make z increasing order
    #print(SMF_z_list.shape, SMF_Phi_list.shape)
    # 
    if z is None:
        lgPhiMstar_matrix = np.log10(SMF_Phi_list.T)
        return SMF_z_list, lgMstar_grid, lgPhiMstar_matrix
    elif z_is_vector:
        #print('calc_SMF_dzliu2018: np.min(SMF_z_list), np.max(SMF_z_list), z:', np.min(SMF_z_list), np.max(SMF_z_list), z)
        lgPhiMstar_grid = interp1d(SMF_z_list, np.log10(SMF_Phi_list), kind='cubic')(z) # kind='nearest' 'linear' 'quadratic' 'cubic'
    else:
        #print('calc_SMF_dzliu2018: np.min(SMF_z_list), np.max(SMF_z_list), z:', np.min(SMF_z_list), np.max(SMF_z_list), z)
        lgPhiMstar_grid = interp1d(SMF_z_list, np.log10(SMF_Phi_list), kind='cubic')(z) # kind='nearest' 'linear' 'quadratic' 'cubic'
    # 
    if lgMstar is None:
        return lgMstar_grid, lgPhiMstar_grid
    else:
        return lgPhiMstar_grid






def calc_SMF_dzliu2020(z=None, lgMstar=None, galaxy_type='SFG', z_list=None, tuning_params='', verbose=True):
    # 
    # dzliu 2020 - optimized choice
    # IMF: Chabrier 2003
    # Outputs: lgMstar_grid, lgPhiMstar_grid
    # 
    # tuning_params: for example, 'D17-no-renorm;'
    # 
    # 
    # check z
    if z is not None:
        if not np.isscalar(z):
            #raise ValueError('Please input a float number as the redshift!')
            if type(z) is list:
                z = np.array(z)
            z_is_vector = True
        else:
            z_is_vector = False
    # 
    # check galaxy_type
    if not (type(galaxy_type) is str):
        raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    else:
        if not (galaxy_type in ['SFG', 'QG', 'ALL']):
            raise ValueError('Please input either "ALL", "SFG" or "QG" as the galaxy_type!')
    # 
    # make lgMstar
    if lgMstar is None:
        lgMstar_grid = np.linspace(6.0, 13.0, num=1000, endpoint=True)
    else:
        lgMstar_grid = lgMstar
    # 
    # code from '/Volumes/GoogleDrive/Team Drives/DeepFields/Simulations/Cosmological_Galaxy_Modelling/a_dzliu_code_Plot_SMF_dzliu_model.sm'
    # make z_list
    if z_list is None:
        ##z_list = np.arange(10.75, 0.75, -0.5).tolist()
        ##z_list.extend([0.75, 0.50, 0.25, 0.125, 0.0625, 0.03125, 0.015625, 0.00000])
        #z_list = np.arange(10.75, 1.00, -0.5).tolist()
        #z_list.extend([1.00, 0.75, 0.50, 0.25, 0.125, 0.0625, 0.03125, 0.015625, 0.00000])
        #print('z_list', z_list, '(len=%d)'%(len(z_list)))
        # 
        # <20200105> precision issue
        # <20200105> now we increase the sampling of z_list from 29 to 40, so that the SMF are better interpolated.
        z_list = [] 
        z_list.extend(np.arange(10.75, 5.00, -0.5).tolist())
        z_list.extend(np.arange(5.00, 1.00, -0.25).tolist())
        z_list.extend(np.arange(1.00, 0.125, -0.125).tolist())
        z_list.extend([0.125, 0.0625, 0.03125, 0.015625, 0.00000])
        #print('z_list', z_list, '(len=%d)'%(len(z_list)))
        # <20200105> precision issue (end)
        # 
        #z_list = []
        #z_list.extend(np.arange(9.75, 4.00, -0.5).tolist())
        #z_list.extend(np.arange(4.00, 3.00, -0.25).tolist())
        #z_list.extend(np.arange(3.00, 2.00, -0.125).tolist())
        #z_list.extend(np.arange(2.00, 1.00, -0.0625).tolist())
        #z_list.extend(np.arange(1.00, 0.00, -0.03125).tolist())
        #z_list.extend([0.00])
    SMF_z_list = []
    SMF_Phi_list = []
    Mstar_cumulated_list = calc_Mstar_integrating_CSFRD_dzliu2018(z_list)
    for i in range(len(z_list)-1):
        z_bin = (z_list[i] + z_list[i+1]) / 2.0
        t_bin = cosmo.age(z_bin).value # Gyr since the Big Bang
        Schechter_M = 10**lgMstar_grid
        Schechter_Mstep = lgMstar_grid[1] - lgMstar_grid[0]
        do_renorm_by_CSFRD_cumulated_Mstar = True
        # 
        # first construct SMF at each redshift bin
        #if False:
        #    EEE_Mchar_MQG = lg((10**10.68)*(t_bin/cosmo.age(0).value)**(0.50)) # Quiescent galaxies' SMF's first component
        #    EEE_Mchar_NQG = lg((10**10.68)*(t_bin/cosmo.age(0).value)**(0.50)) # Quiescent galaxies' SMF's second component
        #    EEE_Mchar_SFG = lg((10**10.72)*(t_bin/cosmo.age(0).value)**(-0.50)) # Star-forming galaxies' SMF
        #    EEE_Phi_MQG = (3.400e-3)*(t_bin/cosmo.age(0).value)**(4.00)
        #    EEE_Phi_NQG = (0.126e-3)*(t_bin/cosmo.age(0).value)**(4.00) # decrease much faster with increasing z
        #    EEE_Phi_SFG = (0.900e-3)*(t_bin/cosmo.age(0).value)**(0.20)
        #    EEE_alpha_MQG = -0.39 + (z_bin)*(1.00)
        #    EEE_alpha_NQG = -1.56 + (z_bin)*(1.00)
        #    EEE_alpha_SFG = -1.40 + (z_bin)*(-0.06) # slope -- this makes too many low mass galaxiest at z>4 -- faint-end of the 24um number counts -- no, this is because stellar SED has too old age, f24um is underestimated!
        #    #EEE_alpha_SFG = -1.40 + (z)*(-0.03) # slope
        #    Schechter_P_MQG = Schechter_Function(lgMstar_grid, EEE_Phi_MQG, EEE_Mchar_MQG, EEE_alpha_MQG)
        #    Schechter_P_NQG = Schechter_Function(lgMstar_grid, EEE_Phi_NQG, EEE_Mchar_NQG, EEE_alpha_NQG)
        #    Schechter_P_SFG = Schechter_Function(lgMstar_grid, EEE_Phi_SFG, EEE_Mchar_SFG, EEE_alpha_SFG)
        #    Schechter_P_QG = Schechter_P_MQG + Schechter_P_NQG
        #    #print('z = %.04f, lgSchechter_P_SFG = %s, lgSchechter_P_QG = %s'%(z_bin, np.log10(Schechter_P_SFG), np.log10(Schechter_P_QG)))
        # 
        if z_bin < 0.025:
            Schechter_P_QG = 10**(calc_SMF_Kelvin2014(0.025, lgMstar=lgMstar_grid, galaxy_type='QG'))
            Schechter_P_SFG = 10**(calc_SMF_Kelvin2014(0.025, lgMstar=lgMstar_grid, galaxy_type='SFG'))
        elif z_bin < 0.06:
            Schechter_P_QG = 10**(calc_SMF_Kelvin2014(z_bin, lgMstar=lgMstar_grid, galaxy_type='QG'))
            Schechter_P_SFG = 10**(calc_SMF_Kelvin2014(z_bin, lgMstar=lgMstar_grid, galaxy_type='SFG'))
        elif z_bin < 0.2:
            # interpolate Kelvin2014 local SMF and Moutard2016 SMF
            t_lower_opz = np.log10(1.0+0.06)
            t_upper_opz = np.log10(1.0+0.2)
            t_opz = np.log10(1.0+z_bin)
            t_lower_factor = (t_upper_opz-t_opz)/(t_upper_opz-t_lower_opz)
            t_upper_factor = 1.0-t_lower_factor
            Schechter_P_QG = 10**(calc_SMF_Kelvin2014(0.06, lgMstar=lgMstar_grid, galaxy_type='QG')) * t_lower_factor + 10**(calc_SMF_Moutard2016(0.2, lgMstar=lgMstar_grid, galaxy_type='QG')) * t_upper_factor
            Schechter_P_SFG = 10**(calc_SMF_Kelvin2014(0.06, lgMstar=lgMstar_grid, galaxy_type='SFG')) * t_lower_factor + 10**(calc_SMF_Moutard2016(0.2, lgMstar=lgMstar_grid, galaxy_type='SFG')) * t_upper_factor
        elif z_bin < 1.5:
            Schechter_P_QG = 10**(calc_SMF_Moutard2016(z_bin, lgMstar=lgMstar_grid, galaxy_type='QG'))
            Schechter_P_SFG = 10**(calc_SMF_Moutard2016(z_bin, lgMstar=lgMstar_grid, galaxy_type='SFG'))
        elif z_bin < 4.0:
            Schechter_P_QG = 10**(calc_SMF_Davidzon2017(z_bin, lgMstar=lgMstar_grid, galaxy_type='QG'))
            Schechter_P_SFG = 10**(calc_SMF_Davidzon2017(z_bin, lgMstar=lgMstar_grid, galaxy_type='SFG'))
            #<TODO># QG/SFG fraction seems lower at z~1-3?
            #if z_bin > 1.0 and z_bin < 3.0:
            #    Schechter_P_QG = Schechter_P_QG * np.interp(z_bin, [1.0, 1.5, 2.0, 2.5, 3.0], [1.0, 1.2, 1.2, 1.2, 1.0])
            if tuning_params is not None:
                if tuning_params.find('D17-no-renorm')>=0:
                    do_renorm_by_CSFRD_cumulated_Mstar = False
        else:
            Schechter_P_QG = 10**(calc_SMF_Davidzon2017(4.0, lgMstar=lgMstar_grid, galaxy_type='QG'))
            Schechter_P_SFG = 10**(calc_SMF_Davidzon2017(4.0, lgMstar=lgMstar_grid, galaxy_type='SFG'))
            # 
            # Note that my Schechter_Function already contains * ln(10), i.e., it is per dex!
            # 
        # 
        # 
        # then re-normalize SMF to total stellar mass integrated from CSFRD (assumed some mass loss, see the called function)
        #Mstar_cumulated = calc_Mstar_integrating_CSFRD_dzliu2018(z_bin)
        #Mstar_cumulated = calc_Mstar_integrating_CSFRD_dzliu2018(z_list[i+1]) #<20190915># BUGGY, should be z bin edge, and z_list is in descending order 
        Mstar_cumulated = Mstar_cumulated_list[i+1]
        #CSFRD_at_z_bin = calc_MadauDickinson2014_CSFRD(z_bin)
        #<TODO><20191001># mask = (Schechter_M>=1e9) #<20191001># to match the Madau & Dickinson (2014) integration limit of '0.03 * L_characteristic', i.e., 0.03 * M_characteristic = 0.03 * 10**10.5 = 
        #<TODO><20191001># Schechter_M_total = sum((Schechter_P_QG[mask]+Schechter_P_SFG[mask])*Schechter_M[mask]*Schechter_Mstep) # P is per dex, but our Schechter_Function already contains ln(10), hence we do not need to multiply ln(10) here. \int P(M) dM = \int P(M) M dln(M) = \int P(M)*ln(10) M dlg(M)
        
        if do_renorm_by_CSFRD_cumulated_Mstar:
            Schechter_M_total = sum((Schechter_P_QG+Schechter_P_SFG)*Schechter_M*Schechter_Mstep) # P is per dex, but our Schechter_Function already contains ln(10), hence we do not need to multiply ln(10) here. \int P(M) dM = \int P(M) M dln(M) = \int P(M)*ln(10) M dlg(M)
            renorm_factor = Mstar_cumulated / Schechter_M_total
            Schechter_P_SFG = Schechter_P_SFG * renorm_factor # per dex but already contains ln(10)
            Schechter_P_QG = Schechter_P_QG * renorm_factor # per dex but already contains ln(10)
        Schechter_P_ALL = Schechter_P_SFG + Schechter_P_QG
        if verbose:
            print('z = %.04f, lgMstar_CSFRD = %0.2f, lgMstar_SMF = %0.2f, renorm = %s'%(z_bin, np.log10(Mstar_cumulated), np.log10(Schechter_M_total), renorm_factor))
        #print('z = %.04f, lgCSFRD = %0.2f, lgMstar_CSFRD = %0.2f, lgMstar_SMF = %0.2f'%(z_bin, np.log10(CSFRD_at_z_bin), np.log10(Mstar_cumulated), np.log10(Schechter_M_total)))
        # 
        # --> checked OK
        # z = 0.0156, lgCSFRD = -2.04, lgMstar_CSFRD = 8.40, lgMstar_SMF = 8.35
        # 
        # 
        SMF_z_list.append(z_list[i+1]) # append the lower redshift end point of each bin
        if galaxy_type == 'SFG':
            SMF_Phi_list.append(Schechter_P_SFG)
        elif galaxy_type == 'QG':
            SMF_Phi_list.append(Schechter_P_QG)
        elif galaxy_type == 'ALL':
            SMF_Phi_list.append(Schechter_P_ALL)
        # 
    # spline at z for the output
    SMF_z_list = np.array(SMF_z_list)[::-1] # make z increasing order
    SMF_Phi_list = np.array(SMF_Phi_list)[::-1].T # make z increasing order
    #print(SMF_z_list.shape, SMF_Phi_list.shape)
    # 
    if z is None:
        lgPhiMstar_matrix = np.log10(SMF_Phi_list.T)
        return SMF_z_list, lgMstar_grid, lgPhiMstar_matrix
    elif z_is_vector:
        #print('calc_SMF_dzliu2018: np.min(SMF_z_list), np.max(SMF_z_list), z:', np.min(SMF_z_list), np.max(SMF_z_list), z)
        lgPhiMstar_grid = interp1d(SMF_z_list, np.log10(SMF_Phi_list), kind='cubic')(z) # kind='nearest' 'linear' 'quadratic' 'cubic'
    else:
        #print('calc_SMF_dzliu2018: np.min(SMF_z_list), np.max(SMF_z_list), z:', np.min(SMF_z_list), np.max(SMF_z_list), z)
        lgPhiMstar_grid = interp1d(SMF_z_list, np.log10(SMF_Phi_list), kind='cubic')(z) # kind='nearest' 'linear' 'quadratic' 'cubic'
    # 
    if lgMstar is None:
        return lgMstar_grid, lgPhiMstar_grid
    else:
        return lgPhiMstar_grid






















#!/usr/bin/env python
# 
# Usage:
#    from calc_fmol import ( calc_fmol_from_metalZ_following_Krumholz2009, 
#                            calc_fmol_from_metalZ_following_Dave2016, 
#                            calc_fmol_from_metalZ_following_Popping2014 )
# 
# 

from __future__ import print_function

import os, sys, re, json, time, astropy
import numpy as np
from astropy.table import Table, Column, hstack
from astropy import units as u
from astropy import constants as const
from astropy.modeling.blackbody import blackbody_lambda, blackbody_nu
from copy import copy

from astropy.cosmology import FlatLambdaCDM
cosmo = FlatLambdaCDM(H0=70, Om0=0.27, Tcmb0=2.725)

if sys.version_info.major >= 3:
    long = int
else:
    pass



# 
# def functions
# these functions are copied from 'AlmaCosmos/Pipeline/a3cosmos-Sample-Selection/a_dzliu_code_step_8_output_datatable_ID_RA_Dec_z_Mstar_SFR_sSFR_with_deltaGas.py'
# 
def calc_Lv_from_flux_density(obs_flux_mJy, obs_wavelength_um, redshift):
    lumdist_Mpc = cosmo.luminosity_distance(redshift).value # Mpc
    Lv = 4 * np.pi * lumdist_Mpc**2 * obs_flux_mJy / (1.+redshift) / 40.31970 / 1e9 # Lsun Hz-1 rest-frame, using 1 Lsun Mpc^-2 = 40.31970 mJy GHz.
    Lv = Lv * 3.839e33 # erg s-1 Hz-1
    return Lv

def calc_Lv_850um_from_RF_850um(RF_850um_mJy, redshift):
    lumdist_Mpc = cosmo.luminosity_distance(redshift).value # Mpc
    vLv_850um = 4 * np.pi * lumdist_Mpc**2 * RF_850um_mJy / 40.31970 * (2.99792458e5/(850.0)) # Lsun, using 1 Lsun Mpc^-2 = 40.31970 mJy GHz.
    Lv_850um = vLv_850um * 3.839e33 / ((2.99792458e5/850.0)*1e9) # erg s-1 Hz-1
    Lv_850um_2 = 1.19e27 * (RF_850um_mJy/1e3) / (1.+redshift) * lumdist_Mpc**2 # erg s-1 Hz-1, see Hughes 2017 - 1702.07350.pdf (page 3 right)
    print(Lv_850um, Lv_850um_2) # checked that `calc_Lv_850um_from_RF_850um(1., 2.0)` gives "1.0156146866563148e+32 1.0101020437871695e+32"
    return Lv_850um

def calc_M_mol_Hughes2017(S_850um_mJy, redshift):
    # Hughes+2017 
    # Here we take their molecular gas mass calibration. They also provided M_ISM calibration. 
    # For massive (log (vLv / [erg s-1 Hz-1]) ~ 32) galaxies, M_ISM - M_mol = 0.01 dex.
    # For low mass (log (vLv / [erg s-1 Hz-1]) ~ 30) galaxies, M_ISM - M_mol = 0.13 dex.
    # 
    Lv_850um_erg_s_Hz = calc_Lv_from_flux_density(S_850um_mJy, 850.0*(1.+redshift), redshift)
    M_mol_Msun = np.power(10.0, np.log10(Lv_850um_erg_s_Hz) * 0.93 - 17.74) # Hughes 2017 - 1702.07350.pdf (page 4, equation 5)
    return M_mol_Msun

def calc_M_total_gas_Hughes2017(S_850um_mJy, redshift):
    # Hughes+2017 
    # Here we take their molecular gas mass calibration. They also provided M_ISM calibration. 
    # For massive (log (vLv / [erg s-1 Hz-1]) ~ 32) galaxies, M_ISM - M_mol = 0.01 dex.
    # For low mass (log (vLv / [erg s-1 Hz-1]) ~ 30) galaxies, M_ISM - M_mol = 0.13 dex.
    # 
    Lv_850um_erg_s_Hz = calc_Lv_from_flux_density(S_850um_mJy, 850.0*(1.+redshift), redshift)
    M_total_gas_Msun = np.power(10.0, np.log10(Lv_850um_erg_s_Hz) * 0.86 - 15.38) # Hughes 2017 - 1702.07350.pdf (page 4, equation 4, for total ISM mass)
    return M_total_gas_Msun

def calc_M_mol_Groves2015(obs_flux_mJy, obs_wavelength_um, redshift):
    # Groves+2015
    # They gas mass calibration are for total atomic+mol gas, 
    # but they also mentioned that for massive galaxies M_mol/M_ISM ~ 1, see their last paragraph in Sect. 3.3. 
    # 
    #Brent_W = np.array([250., 350., 500.])
    #Brent_A = np.array([1.57, 1.49, 1.72]) # see Groves, bibcode=2015ApJ...799...96G, Table 6
    #Brent_B = np.array([0.86, 0.92, 0.96]) # see Groves, bibcode=2015ApJ...799...96G, Table 6, log(Mgas) = A + B * log(nuLnu_W)
    Brent_W = np.array([70.0, 100., 160., 250., 350., 500.])
    Brent_A = np.array([4.60, 3.27, 1.91, 1.17, 1.17, 1.44]) # see Groves, bibcode=2015ApJ...799...96G, Table 5, for log(M⋆/M⊙) > 9 sample
    Brent_B = np.array([0.50, 0.63, 0.78, 0.90, 0.90, 0.99]) # see Groves, bibcode=2015ApJ...799...96G, Table 5, for log(M⋆/M⊙) > 9 sample, log(Mgas) = A + B * log(nuLnu_W)
    rest_wavelength_um = obs_wavelength_um / (1.+redshift)
    dL = cosmo.luminosity_distance(redshift).value # Mpc
    Brent_log_M_mol_Msun = np.zeros((len(obs_flux_mJy),len(Brent_W),))
    #print(Brent_log_M_mol_Msun)
    for i in range(len(Brent_W)):
        A = Brent_A[i]
        B = Brent_B[i]
        vLv = (2.99792458e5/obs_wavelength_um)*(obs_flux_mJy)*(4*np.pi*dL**2)/40.31970 # "/40.31970" converts mJy GHz to Lsun Mpc-2 # 1.19e27 * Snu / (1+z) * dL**2 # erg s-1 Hz-1, see Hughes 2017 - 1702.07350.pdf (page 3 right)
        log_M_ISM_Msun = A + B * np.log10(vLv) # Groves et al. 2015, page 6, right middle-to-bottom
        Brent_log_M_mol_Msun[:,i] = log_M_ISM_Msun # calculate one M_mol for each Brent parameter, here we assume M_ISM is M_mol for massive galaxies.
    # 
    log_M_mol_Msun = []
    for i in range(len(obs_flux_mJy)):
        if rest_wavelength_um[i] >= np.min(Brent_W) and rest_wavelength_um[i] <= np.max(Brent_W):
            Brent_Interp = scipy.interpolate.interp1d(np.log10(Brent_W), Brent_log_M_mol_Msun[i,:])
            log_M_mol_Msun.append(Brent_Interp(np.log10(rest_wavelength_um[i])))
        else:
            log_M_mol_Msun.append(np.nan)
    M_mol_Msun = 10**np.array(log_M_mol_Msun)
    # 
    Schinnerer2016_A = 1.57 - 8e-4 * (rest_wavelength_um-250.0)
    Schinnerer2016_B = 0.86 + 6e-4 * (rest_wavelength_um-250.0)
    lgMH2_Schinnerer2016 = Schinnerer2016_A + Schinnerer2016_B * np.log10(vLv) # Schinnerer et al. 2016, equation (1)
    # 
    return M_mol_Msun

def calc_M_total_gas_Groves2015(obs_flux_mJy, obs_wavelength_um, redshift):
    # Groves+2015
    # They gas mass calibration are for total atomic+mol gas, 
    # but they also mentioned that for massive galaxies M_mol/M_ISM ~ 1, see their last paragraph in Sect. 3.3. 
    # 
    #Brent_W = np.array([250., 350., 500.])
    #Brent_A = np.array([1.57, 1.49, 1.72]) # see Groves, bibcode=2015ApJ...799...96G, Table 6
    #Brent_B = np.array([0.86, 0.92, 0.96]) # see Groves, bibcode=2015ApJ...799...96G, Table 6, log(Mgas) = A + B * log(nuLnu_W)
    Brent_W = np.array([70.0, 100., 160., 250., 350., 500.])
    Brent_A = np.array([4.15, 3.62, 3.52, 3.17, 3.08, 3.19]) # see Groves, bibcode=2015ApJ...799...96G, Table 5, for all log(M⋆/M⊙) sample
    Brent_B = np.array([0.55, 0.60, 0.61, 0.69, 0.74, 0.78]) # see Groves, bibcode=2015ApJ...799...96G, Table 5, for all log(M⋆/M⊙) sample, log(Mgas) = A + B * log(nuLnu_W)
    rest_wavelength_um = obs_wavelength_um / (1.+redshift)
    dL = cosmo.luminosity_distance(redshift).value # Mpc
    Brent_log_M_mol_Msun = np.zeros((len(obs_flux_mJy),len(Brent_W),))
    #print(Brent_log_M_mol_Msun)
    for i in range(len(Brent_W)):
        A = Brent_A[i]
        B = Brent_B[i]
        vLv = (2.99792458e5/obs_wavelength_um)*(obs_flux_mJy)*(4*np.pi*dL**2)/40.31970 # "/40.31970" converts mJy GHz to Lsun Mpc-2 # 1.19e27 * Snu / (1+z) * dL**2 # erg s-1 Hz-1, see Hughes 2017 - 1702.07350.pdf (page 3 right)
        log_M_ISM_Msun = A + B * np.log10(vLv) # Groves et al. 2015, page 6, right middle-to-bottom
        Brent_log_M_mol_Msun[:,i] = log_M_ISM_Msun # calculate one M_mol for each Brent parameter, here we assume M_ISM is M_mol for massive galaxies.
    # 
    log_M_mol_Msun = []
    for i in range(len(obs_flux_mJy)):
        if rest_wavelength_um[i] >= np.min(Brent_W) and rest_wavelength_um[i] <= np.max(Brent_W):
            Brent_Interp = scipy.interpolate.interp1d(np.log10(Brent_W), Brent_log_M_mol_Msun[i,:]) # interpolate the coefficients <TODO>
            log_M_mol_Msun.append(Brent_Interp(np.log10(rest_wavelength_um[i])))
        else:
            log_M_mol_Msun.append(np.nan)
    M_mol_Msun = 10**np.array(log_M_mol_Msun)
    # 
    return M_mol_Msun

def calc_M_mol_Scoville2017_RF(S_850um_mJy, redshift):
    # As mentioned in Hughes+2017, Scoville+2014 calibrated the atomic+mol gas mass, 
    #   but Scoville+2016 and 2017 calibrated only the mol gas mass.
    # 
    Lv_850um_erg_s_Hz = calc_Lv_from_flux_density(S_850um_mJy, 850.0*(1.+redshift), redshift)
    M_mol_Msun = Lv_850um_erg_s_Hz / 6.7e19 # Scoville et al. 2016ApJ...820...83S (appendix and also page 16 equation 16) -- molecular gas including Helium
    #lumdist_Mpc = cosmo.luminosity_distance(redshift).value # Mpc
    #M_mol_Msun_2 = 1.78 * RF_850um_mJy/(1.+redshift) * (lumdist_Mpc/1e3)**2 * 1e10 # Scoville et al. 2016ApJ...820...83S (appendix and also page 16 equation 16) -- molecular gas including Helium
    #print('%0.6e %0.6e' % (M_mol_Msun, M_mol_Msun_2) ) # checked that `calc_M_mol_Scoville2017(1., 2.0)` "1.515843e+12 1.510909e+12"
    return M_mol_Msun

def calc_M_mol_Scoville2017(obs_flux_mJy, obs_wavelength_um, redshift):
    # As mentioned in Hughes+2017, Scoville+2014 calibrated the atomic+mol gas mass, 
    #   but Scoville+2016 and 2017 calibrated only the mol gas mass.
    # 
    Lv_obs_erg_s_Hz = calc_Lv_from_flux_density(obs_flux_mJy, obs_wavelength_um, redshift)
    beta = 1.8 # do K-correction assuming planck function, beta=1.8, Tdust=25K
    Tdust = 25.0 # K
    cal_Lv_obs_erg_s_Hz = blackbody_nu(obs_wavelength_um/(1.+redshift) * u.um, Tdust * u.K) * np.power(2.99792458e5/(obs_wavelength_um/(1.+redshift)), beta)
    cal_Lv_850um_erg_s_Hz = blackbody_nu(850.0 * u.um, 25.0 * u.K) * np.power(2.99792458e5/850.0, beta)
    #print(cal_Lv_obs_erg_s_Hz)
    #print(cal_Lv_850um_erg_s_Hz)
    #print(cal_Lv_obs_erg_s_Hz / cal_Lv_850um_erg_s_Hz)
    Lv_850um_erg_s_Hz = Lv_obs_erg_s_Hz * (cal_Lv_850um_erg_s_Hz/cal_Lv_obs_erg_s_Hz)
    M_mol_Msun = Lv_850um_erg_s_Hz / 6.7e19 # Scoville et al. 2016ApJ...820...83S (appendix and also page 16 equation 16) -- molecular gas including Helium
    return M_mol_Msun



# 
# main function 1
# 
def calc_gas_mass_from_dust_continuum(obs_wavelength_um, obs_flux_mJy, SED_flux_at_obs_wavelength = None, SED_flux_at_rest_850um = None, z = None, method = 'Hughes2017'):
    # 
    # check user input
    # 
    if obs_wavelength_um is None:
        raise ValueError('Error! Please input \'obs_wavelength_um\' for calc_gas_mass_from_dust_continuum_with_band_conversion()!')
    # 
    if obs_flux_mJy is None:
        raise ValueError('Error! Please input \'obs_flux_mJy\' for calc_gas_mass_from_dust_continuum_with_band_conversion()!')
    # 
    if z is None:
        raise ValueError('Error! Please input \'z\' for calc_gas_mass_from_dust_continuum_with_band_conversion()!')
    # 
    # convert observed dust continuum at observed wavelength to (cold molecular) gas mass using the input method. 
    # 
    allowed_methods = ['Hughes2017', 'Groves2015', 'Scoville2017']
    # 
    if method == 'Hughes2017':
        # 
        if SED_flux_at_obs_wavelength is None or SED_flux_at_rest_850um is None:
            raise ValueError('Error! Please input \'SED_flux_at_obs_wavelength\' and \'SED_flux_at_rest_850um\' for calc_gas_mass_from_dust_continuum_with_band_conversion() with method \"Hughes2017\"!')
        # 
        M_mol_gas = calc_M_mol_Hughes2017(SED_flux_at_rest_850um / SED_flux_at_obs_wavelength * obs_flux_mJy, z) # band conversion used MAGPHYS SED, then alphamol850 following Hughes+2017.
        # 
    elif method == 'Groves2015':
        # 
        M_mol_gas = calc_M_mol_Groves2015(obs_flux_mJy, obs_wavelength_um, z) # no need band conversion here, it is inside the called function.
        # 
    elif method == 'Scoville2017':
        # 
        M_mol_gas = calc_M_mol_Scoville2017(obs_flux_mJy, obs_wavelength_um, z) # no need band conversion here, it is inside the called function by assuming a simple dust RJ tail slope exactly following Scoville+2017.
        # 
    else:
        # 
        raise NotImplementedError('Sorry! The gas mass calibration has not been implemented for the method "%s"! We have following methods: %s'%(method, allowed_methods))
        # 
    # 
    return M_mol_gas, method




# 
# def functions
# 
def calc_molecular_hydrogen_fraction(metalZOH, U_MW = None, method = 'Krumholz2009'):
    # 
    script_dir = os.path.dirname(os.path.abspath(__file__))
    if not script_dir in sys.path:
        sys.path.append(script_dir)
    import inspect
    # 
    allowed_methods = ['Krumholz2009', 'Dave2016', 'Popping2014']
    # 
    if re.match(r'.*\bKrumholz2009\b.*', method, re.IGNORECASE):
        # 
        if not ('calc_fmol_from_metalZ_following_Krumholz2009' in dir()):
            from calc_fmol import calc_fmol_from_metalZ_following_Krumholz2009
        # 
        f_mol = calc_fmol_from_metalZ_following_Krumholz2009(metalZOH)
        # 
    elif re.match(r'.*\bDave2016\b.*', method, re.IGNORECASE):
        # 
        if not ('calc_fmol_from_metalZ_following_Dave2016' in dir()):
            from calc_fmol import calc_fmol_from_metalZ_following_Dave2016
        # 
        f_mol = calc_fmol_from_metalZ_following_Dave2016(metalZOH)
        # 
    elif re.match(r'.*\bPopping2014\b.*', method, re.IGNORECASE):
        # 
        if not ('calc_fmol_from_metalZ_following_Popping2014' in dir()):
            from calc_fmol import calc_fmol_from_metalZ_following_Popping2014
        # 
        f_mol = calc_fmol_from_metalZ_following_Popping2014(metalZOH, U_MW = U_MW)
        # 
    else:
        # 
        raise NotImplementedError('Sorry! The molecular hydrogen fraction calculation has not been implemented for the method "%s"! We have following methods: %s'%(method, allowed_methods))
        # 
    return f_mol
    


# 
# main function 2
# 
def calc_gas_mass_from_dust_mass(M_dust, M_star = None, SFR = None, metallicity = None, GDR = None, z = None, method = ''):
    # 
    script_dir = os.path.dirname(os.path.abspath(__file__))
    if not script_dir in sys.path:
        sys.path.append(script_dir)
    #<TETS>#import inspect
    #<TETS>#print('inspect.isfunction(calc_deltaGD_from_metalZ_following_Genzel2015)', inspect.isfunction('calc_deltaGD_from_metalZ_following_Genzel2015'))
    #<TETS>#print('in dir()', 'calc_deltaGD_from_metalZ_following_Genzel2015' in dir())
    #<TETS>#from calc_delta_GD import calc_deltaGD_from_metalZ_following_Genzel2015
    #<TETS>#print('inspect.isfunction(calc_deltaGD_from_metalZ_following_Genzel2015)', inspect.isfunction('calc_deltaGD_from_metalZ_following_Genzel2015'))
    #<TETS>#print('in dir()', 'calc_deltaGD_from_metalZ_following_Genzel2015' in dir())
    # 
    M_total_gas = None
    M_mol_gas = None
    # 
    allowed_methods = ['Magdis2012', 'Genzel2015a', 'Genzel2015b', 'GDR']
    # 
    # if method is not given by the user
    if method == '' and GDR is not None:
        method = 'GDR'
    elif method == '':
        method = 'default'
    # 
    if re.match(r'.*\bMagdis2012\b.*', method, re.IGNORECASE):
        # 
        if metallicity is None:
            if M_star is None or SFR is None:
                raise ValueError('Error! Please input \'M_star\' and \'SFR\' or \'metallicity\' for calc_gas_mass_from_dust_mass() with the method \"Magdis2012\"!')
            else:
                if not ('calc_metalZ_from_FMR_following_Mannucci2010' in dir()):
                    from calc_metal_Z import calc_metalZ_from_FMR_following_Mannucci2010
                metallicity = calc_metalZ_from_FMR_following_Mannucci2010(M_star, SFR)
                method += ', lgMstar=%.2f, SFR=%g, [Mannucci2010] metallicity=%.2f'%(np.log10(M_star), SFR, metallicity)
        else:
            method += ', metallicity=%.2f'%(metallicity)
        # 
        if GDR is None:
            if not ('calc_deltaGD_from_metalZ_following_Magdis2012' in dir()):
                from calc_delta_GD import calc_deltaGD_from_metalZ_following_Magdis2012
            GDR = calc_deltaGD_from_metalZ_following_Magdis2012(metallicity)
            method += ', [Magdis2012] GDR=%.2f'%(GDR)
        else:
            # 
            method += ', GDR=%.2f'%(GDR)
        # 
        M_mol_gas = M_dust * GDR
        # 
        # if method is like 'Magdis2012+Krumholz2009'
        if method.find('+') >= 0:
            f_mol = calc_molecular_hydrogen_fraction(metallicity, method = method)
            M_total_gas = M_mol_gas
            M_mol_gas = M_total_gas * f_mol
        # 
    elif re.match(r'.*\bGenzel2015a\b.*', method, re.IGNORECASE):
        # 
        if metallicity is None:
            if M_star is None or SFR is None or z is None:
                raise ValueError('Error! Please input (\'M_star\' and \'SFR\' and \'z\') or \'metallicity\' for calc_gas_mass_from_dust_mass() with the method \"Genzel2015a\"!')
            else:
                if not ('calc_metalZ_from_FMR_following_Genzel2015a' in dir()):
                    from calc_metal_Z import calc_metalZ_from_FMR_following_Genzel2015a
                metallicity = calc_metalZ_from_FMR_following_Genzel2015a(M_star, SFR, z)
                method += ', lgMstar=%.2f, SFR=%g, z=%g, [Genzel2015a] metallicity=%.2f'%(np.log10(M_star), SFR, z, metallicity)
        else:
            method += ', metallicity=%.2f'%(metallicity)
        # 
        if GDR is None:
            if not ('calc_deltaGD_from_metalZ_following_Genzel2015' in dir()):
                from calc_delta_GD import calc_deltaGD_from_metalZ_following_Genzel2015
            GDR = calc_deltaGD_from_metalZ_following_Genzel2015(metallicity)
            method += ', [Genzel2015] GDR=%.2f'%(GDR)
        else:
            # 
            method += ', GDR=%.2f'%(GDR)
        # 
        M_mol_gas = M_dust * GDR
        # 
        # if method is like 'Genzel2015a+Krumholz2009'
        if method.find('+') >= 0:
            f_mol = calc_molecular_hydrogen_fraction(metallicity, method = method)
            M_total_gas = M_mol_gas
            M_mol_gas = M_total_gas * f_mol
        # 
    elif re.match(r'.*\b(Genzel2015b|Genzel2015)\b.*', method, re.IGNORECASE):
        # 
        if metallicity is None:
            if M_star is None or SFR is None:
                raise ValueError('Error! Please input (\'M_star\' and \'SFR\') or \'metallicity\' for calc_gas_mass_from_dust_mass() with the method \"Genzel2015b\"!')
            else:
                if not ('calc_metalZ_from_FMR_following_Genzel2015b' in dir()):
                    from calc_metal_Z import calc_metalZ_from_FMR_following_Genzel2015b
                metallicity = calc_metalZ_from_FMR_following_Genzel2015b(M_star, SFR)
                method += ', lgMstar=%.2f, SFR=%g, [Genzel2015b] metallicity=%.2f'%(np.log10(M_star), SFR, metallicity)
        # 
        if GDR is None:
            if not ('calc_deltaGD_from_metalZ_following_Genzel2015' in dir()):
                from calc_delta_GD import calc_deltaGD_from_metalZ_following_Genzel2015
            GDR = calc_deltaGD_from_metalZ_following_Genzel2015(metallicity)
            method += ', [Genzel2015] GDR=%.2f'%(GDR)
        else:
            # 
            method += ', GDR=%.2f'%(GDR)
        # 
        M_mol_gas = M_dust * GDR
        # 
        # if method is like 'Genzel2015a+Krumholz2009'
        if method.find('+') >= 0:
            f_mol = calc_molecular_hydrogen_fraction(metallicity, method = method)
            M_total_gas = M_mol_gas
            M_mol_gas = M_total_gas * f_mol
        # 
        # 
    elif re.match(r'.*\b(default)\b.*', method, re.IGNORECASE):
        # 
        if metallicity is None:
            if M_star is None or SFR is None or z is None:
                raise ValueError('Error! Please input (\'M_star\' and \'SFR\' and \'z\') or \'metallicity\' for calc_gas_mass_from_dust_mass() with the method \"default\"!')
            else:
                if not ('calc_metalZ_from_FMR_following_Genzel2015ab_combined_by_dzliu' in dir()):
                    from calc_metal_Z import calc_metalZ_from_FMR_following_Genzel2015ab_combined_by_dzliu
                metallicity = calc_metalZ_from_FMR_following_Genzel2015ab_combined_by_dzliu(M_star, SFR, z)
                method += ', lgMstar=%.2f, SFR=%g, [Genzel2015abcomb] metallicity=%.2f'%(np.log10(M_star), SFR, metallicity)
        else:
            method += ', metallicity=%.2f'%(metallicity)
        # 
        if GDR is None:
            if not ('calc_deltaGD_from_metalZ_following_RemyRuyer2014b' in dir()):
                from calc_delta_GD import calc_deltaGD_from_metalZ_following_RemyRuyer2014b
            GDR = calc_deltaGD_from_metalZ_following_RemyRuyer2014b(metallicity)
            method += ', [RemyRuyer2014b] GDR=%.2f'%(GDR)
        else:
            # 
            method += ', GDR=%.2f'%(GDR)
        # 
        M_mol_gas = M_dust * GDR
        # 
        # if method is like 'Genzel2015a+Krumholz2009'
        if method.find('+') >= 0:
            f_mol = calc_molecular_hydrogen_fraction(metallicity, method = method)
            M_total_gas = M_mol_gas
            M_mol_gas = M_total_gas * f_mol
        # 
    elif re.match(r'.*\bGDR\b.*', method, re.IGNORECASE):
        # Assuming a single GDR
        # 
        if GDR is None:
            raise ValueError('Error! Please input \'GDR\' for calc_gas_mass_from_dust_mass() with the method \"GDR\"!')
            #if metallicity is None:
            #    raise ValueError('Error! Please input \'metallicity\' or \'GDR\' for calc_gas_mass_from_dust_mass() with the method \"GDR\"!')
            #else:
            #    if not ('calc_deltaGD_from_metalZ_following_Genzel2015' in dir()):
            #        from calc_delta_GD import calc_deltaGD_from_metalZ_following_Genzel2015
            #    GDR = calc_deltaGD_from_metalZ_following_Genzel2015(metallicity)
        else:
            # 
            method += ', GDR=%.2f'%(GDR)
        # 
        M_mol_gas = M_dust * GDR
        # 
        # if method is like 'Genzel2015a+Krumholz2009'
        if method.find('+') >= 0:
            f_mol = calc_molecular_hydrogen_fraction(metallicity, method = method)
            M_total_gas = M_mol_gas
            M_mol_gas = M_total_gas * f_mol
        # 
    else:
        # 
        raise NotImplementedError('Sorry! The gas mass calibration has not been implemented for the method "%s"! We have following methods: %s'%(method, allowed_methods))
        # 
    # 
    if M_total_gas is None:
        return M_mol_gas, method
    else:
        return M_mol_gas, M_total_gas, method









#!/usr/bin/env python
# 

from __future__ import print_function

import os, sys, re, json, time, astropy
import numpy as np
from astropy.table import Table, Column, hstack
from numpy import log10, power as pow

#from astropy.cosmology import FlatLambdaCDM
##cosmo = FlatLambdaCDM(H0=73, Om0=0.27, Tcmb0=2.725) # replaced since 2019-02-21
#cosmo = FlatLambdaCDM(H0=70, Om0=0.27, Tcmb0=2.725)
if not (os.path.dirname(os.path.abspath(__file__)) in sys.path): sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import apply_cosmology
cosmo = apply_cosmology.cosmo

if sys.version_info.major >= 3:
    long = int
else:
    pass



# 
# def 
# 
def calc_gas_depletion_time_Scoville2017(z, lgMstar=10.5, DeltaMS=0.0):
    # Scoville et al. 2017 (The Astrophysical Journal, 837:150 (20pp), 2017 March 10)
    # Table 2
    Ratio_M_molgas_SFR = 3.23 * (1+z)**(-1.04) * (10**DeltaMS)**(-0.70) * (10**(lgMstar-10.0))**(-0.01)
    return Ratio_M_molgas_SFR

def calc_gas_depletion_time_Tacconi2018(z, lgMstar=10.5, DeltaMS=0.0):
    # Tacconi et al. 2018 (The Astrophysical Journal, 853:179 (22pp), 2018 February 1 https://doi.org/10.3847/1538-4357/aaa4b4)
    # Table 3
    # Best (with bootstrap errors) with S14 (Speagle+2014) MS
    A = +0.09
    B = -0.62
    C = -0.44
    D = +0.09
    E = +0.11
    DeltaRe = 0.0 # log10 (R_e / R_e_0)
    log10_Ratio_M_molgas_SFR = A + B * (np.log10(1+z)) + C * DeltaMS + D * (lgMstar-10.7) + E * (DeltaRe)
    Ratio_M_molgas_SFR = 10**log10_Ratio_M_molgas_SFR
    return Ratio_M_molgas_SFR

def calc_gas_depletion_time_A3COSMOS(z=None, cosmic_age=None, lgMstar=10.5, DeltaMS=0.0):
    # A3COSMOS 
    # Sample 20181203 
    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plot/Plot_z_tauDepl_data_fitting/20190716_with_gas_mass_calibration_H17/fitting_dzliu_function/best_fit_via_MCMC.json'
    #popt = [-0.5820833042317259, -0.5346887707021821, -0.0038760530734522902, 0.041705991802229114, 0.1252905060544558, 0.0592184169424943]
    popt = [-0.5816185208613138, -0.5338752260098323, -0.0038228411346192814, 0.040738011277865915, 0.12494331865991803, 0.05914760529893037]
    a, b, c, d, ak, ck = popt
    if z is None and cosmic_age is None:
        print('Error! Please input either z or cosmic_age')
        sys.exit()
    else:
        if cosmic_age is not None:
            cosmoAge = cosmic_age
        else:
            cosmoAge = cosmo.age(z).value
    log10_Ratio_M_molgas_SFR = (a+ak*(lgMstar-10.0))*DeltaMS + b*(lgMstar-10.0) + (c+ck*(lgMstar-10.0))*(cosmoAge) + d
    Ratio_M_molgas_SFR = 10**log10_Ratio_M_molgas_SFR
    return Ratio_M_molgas_SFR

#def calc_gas_depletion_time_A3COSMOS(z, lgMstar=10.5, DeltaMS=0.0):
#    # A3COSMOS 
#    # Sample 20181203 
#    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plots/Plot_z_tauDepl_data_fitting/best_fit_via_CurveFit.json'
#    #popt = [0.1, -0.4979428049052056, -0.5172364530151924, -0.5, -0.12515630072829534] #<201901>#
#    #popt = [-0.5422, -0.4824, +0.04271, -0.01896, +0.08644] #<20190205>#
#    #popt = [-0.5817, -0.5640, 0.05553, -0.01192, 0.1375] #<20190207># fixed xCOLDGASS Z_PP04_O3N2
#    #popt = [-0.5741, -0.5664, +0.05543, -0.004397, +0.1154] #<20190208># still fitting Bertemes and Lee's data
#    popt = [-0.5702, -0.5741, 0.05489, 0.008601, -0.01334] #<20190301># Eq850 this work, RemyRuyer GDR, Genzel2015_Eq12a_with_dzliu_limit MZR, KMT09 fmol, fitting all available data '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plots/Plot_z_tauDepl_data_fitting/20190301_with_gas_mass_calibration_Eq850_dzliu/best_fit_via_MCMC.json'
#    a, b, c, d, e = popt
#    #log10_Ratio_M_molgas_SFR = b*DeltaMS + c*np.log10(1+z) + e #<201901>#
#    cosmoAge = cosmo.age(z).value
#    log10_Ratio_M_molgas_SFR = a*DeltaMS + (b+c*cosmoAge)*(lgMstar-10.0) + d*(cosmoAge) + e #<20190205>#
#    
#    #<20190714># 
#    a = -0.5788835106242605
#    b = -0.5376145163745489
#    c = -0.004112269290743598
#    d = 0.04527343126448624
#    ak = 0.12203773985943833
#    ck = 0.05930642823001797
#    cosmoAge = cosmo.age(z).value
#    log10_Ratio_M_molgas_SFR = (a+ak*(lgMstar-10.0))*DeltaMS + b*(lgMstar-10.0) + (c+ck*(lgMstar-10.0))*cosmoAge + d
#    
#    
#    Ratio_M_molgas_SFR = 10**log10_Ratio_M_molgas_SFR
#    return Ratio_M_molgas_SFR

#def calc_gas_depletion_time_A3COSMOS_with_dzliu_850_gas_mass_calibration(z, lgMstar=10.5, DeltaMS=0.0):
#    # A3COSMOS 
#    # Sample 20181203 
#    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plots/Plot_z_tauDepl_data_fitting/20190122_with_Leroy_GDR_with_Genzel_Eq12a_MZR_with_dzliu_gas_mass_calibration_with_KMT09_fmol/best_fit_via_CurveFit.json'
#    #popt = [0.1, -0.5191087642422786, -1.5519298330096012, -0.5, 0.08064040034418785] #<20190122_with_Leroy_GDR_with_Genzel_Eq12a_MZR_with_KMT09_fmol>#
#    popt = [-0.5702, -0.5741, 0.05489, 0.008601, -0.01334] #<20190301># Eq850 this work, RemyRuyer GDR, Genzel2015_Eq12a_with_dzliu_limit MZR, KMT09 fmol, fitting all available data '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plots/Plot_z_tauDepl_data_fitting/20190301_with_gas_mass_calibration_Eq850_dzliu/best_fit_via_MCMC.json'
#    a, b, c, d, e = popt
#    log10_Ratio_M_molgas_SFR = b*DeltaMS + c*np.log10(1+z) + e
#    Ratio_M_molgas_SFR = 10**log10_Ratio_M_molgas_SFR
#    return Ratio_M_molgas_SFR

#def calc_gas_depletion_time_A3COSMOS_with_Hughes2017_GasMassCalibration_Leslie2019_MS(z, lgMstar=10.5, DeltaMS=0.0):
#    # A3COSMOS 
#    # Sample 20181203 
#    # MS: Leslie20190515
#    # GasMassCalibration: Hughes2017
#    # fitting '/Users/dzliu/Cloud/GitLab/AlmaCosmos/Plots/Plot_z_tauDepl_data_fitting/20190301_with_gas_mass_calibration_H17_with_MS_Leslie20190515/best_fit_via_MCMC.json'
#    popt = [-0.4367323964697807, -0.31995589347193487, 0.0363679934437835, 0.011782986854565358, -0.19084213551539264]
#    a, b, c, d, e = popt
#    #log10_Ratio_M_molgas_SFR = b*DeltaMS + c*np.log10(1+z) + e #<201901>#
#    cosmoAge = cosmo.age(z).value
#    log10_Ratio_M_molgas_SFR = a*DeltaMS + (b+c*cosmoAge)*(lgMstar-10.0) + d*(cosmoAge) + e #<20190205>#
#    Ratio_M_molgas_SFR = 10**log10_Ratio_M_molgas_SFR
#    return Ratio_M_molgas_SFR






# 
# def core functions (adapted from "a_dzliu_code_fit_z_tauDepl_v6.py")
# 
def func_tauDepl_dzliu_log(pars, a, b, c, d, e): 
    deltaGas,DeltaMS,lgMstar,lgSFR,cosmoAge,z = pars
    # 20181030c
    #tauDepl_model_log = a*DeltaMS + b*(deltaGas) + c*(np.exp(-cosmoAge)) + d
    #tauDepl_model_log = a*(np.log10(deltaGas))*0.0 + b*DeltaMS + c*10**lgSFR*0.0 + lgMstar*0.0 + d*(cosmoAge)*0.0 + e
    tauDepl_model_log = a*(np.log10(deltaGas))*0.0 + b*DeltaMS + 10**lgSFR*0.0 + c*np.log10(1+z) + lgMstar*0.0 + d*(cosmoAge)*0.0 + e
    # 
    return tauDepl_model_log

def func_tauDepl_dzliu(pars, a, b, c, d, e): 
    deltaGas,DeltaMS,lgMstar,lgSFR,cosmoAge,z = pars
    tauDepl_model_log = func_tauDepl_dzliu_log((deltaGas,DeltaMS,lgMstar,lgSFR,cosmoAge,z), a, b, c, d, e)
    tauDepl_model = np.power(10, tauDepl_model_log) # * np.power((1+z),c)
    return tauDepl_model

def func_tauDepl_Tacconi2018_log(pars, a, b, c, d): 
    DeltaMS,lgMstar,z = pars
    tauDepl_model_log = a*DeltaMS + b*(lgMstar-10.7) + c*(np.log10(1+z)) + d
    return tauDepl_model_log

def func_tauDepl_Tacconi2018(pars):
    DeltaMS,lgMstar,z = pars
    tauDepl_model_log = func_tauDepl_Tacconi2018_log((DeltaMS,lgMstar,z), -0.44, 0.09, -0.62, 0.09)
    tauDepl_model = np.power(10, tauDepl_model_log) # * np.power((1+z),c)
    return tauDepl_model

def func_tauDepl_Scoville2017_log(pars, a, b, c, d): 
    DeltaMS,lgMstar,z = pars
    tauDepl_model_log = a*DeltaMS + b*(lgMstar-10.0) + c*(np.log10(1+z)) + d
    return tauDepl_model_log

def func_tauDepl_Scoville2017(pars):
    DeltaMS,lgMstar,z = pars
    tauDepl_model_log = func_tauDepl_Scoville2017_log((DeltaMS,lgMstar,z), -0.70, -0.01, -1.04, np.log10(3.23))
    tauDepl_model = np.power(10, tauDepl_model_log) # * np.power((1+z),c)
    return tauDepl_model












#!/usr/bin/env python
# 

import sys, inspect
from astropy.cosmology import FlatLambdaCDM

def apply_cosmology(Hubble_Constant_z0 = 70, Omega_Matter = 0.27, Omega_Lambda = 0.73, T_CMB_z0 = 2.725):
    # 
    #cosmo = FlatLambdaCDM(H0=73, Om0=0.27, Tcmb0=2.725) # replaced since 2019-02-21
    # 
    # Check out the inspect module, it is used by minimock to mock the caller's scope.
    stack = inspect.stack()
    # 
    # check 'cosmo' variable
    if ('cosmo' in (inspect.stack()[-1][0]).f_globals.keys()):
        cosmo = (inspect.stack()[-1][0]).f_globals['cosmo']
    elif ('cosmo' in (inspect.stack()[-1][0]).f_locals.keys()):
        cosmo = (inspect.stack()[-1][0]).f_locals['cosmo']
    else:
        if '%0.2f'%(Omega_Lambda + Omega_Matter) == '1.00':
            cosmo = FlatLambdaCDM(H0 = Hubble_Constant_z0, Om0 = Omega_Matter, Tcmb0 = T_CMB_z0)
        else:
            raise NotImplementedError('Error! Non-flat cosmology not implemented yet! <TODO>')
    # 
    #print(cosmo.H0)
    # 
    return cosmo



#print((inspect.stack()[-1][0]))
#print((inspect.stack()[-1][0]).f_locals.keys())

#if ('cosmo' in (inspect.stack()[-1][0]).f_globals.keys()):
#    cosmo = (inspect.stack()[-1][0]).f_globals['cosmo']
#elif ('cosmo' in (inspect.stack()[-1][0]).f_locals.keys()):
#    cosmo = (inspect.stack()[-1][0]).f_locals['cosmo']
#else:
#    cosmo = FlatLambdaCDM(H0=70, Om0=0.27, Tcmb0=2.725)

cosmo = apply_cosmology()

#print(cosmo.H0)




#!/usr/bin/env python
# 

from __future__ import print_function

import os, sys, re, json, time, astropy
import numpy as np
from astropy.table import Table, Column, hstack
from copy import copy
from numpy import log, log10, power, sum, sqrt, pi, exp
pow = power
lg = log10
ln = log
from scipy.interpolate import InterpolatedUnivariateSpline, interp1d

if not (os.path.dirname(os.path.abspath(__file__)) in sys.path): sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import apply_cosmology
cosmo = apply_cosmology.cosmo

if sys.version_info.major >= 3:
    long = int
else:
    pass




# 
# def 
# 
def calc_Mmolgas_from_SFR_using_SF_law_Sargent2014(SFR, DeltaMS = 0.0):
    # Sargent et al. 2014, https://ui.adsabs.harvard.edu/abs/2014ApJ...793...19S/abstract
    # bimodal SF law
    arr_alpha = [9.22, 8.05]
    arr_beta = [0.81, 0.81]
    arr_DeltaMS = [0.0, 1.0]
    argmin_DeltaMS = np.argmin(arr_DeltaMS)
    argmax_DeltaMS = np.argmax(arr_DeltaMS)
    min_DeltaMS = arr_DeltaMS[argmin_DeltaMS]
    max_DeltaMS = arr_DeltaMS[argmax_DeltaMS]
    # 
    if np.isscalar(DeltaMS):
        if DeltaMS <= min_DeltaMS:
            t_DeltaMS = min_DeltaMS
            t_alpha = arr_alpha[argmin_DeltaMS]
            t_beta = arr_beta[argmin_DeltaMS]
        elif DeltaMS >= max_DeltaMS:
            t_DeltaMS = max_DeltaMS
            t_alpha = arr_alpha[argmax_DeltaMS]
            t_beta = arr_beta[argmax_DeltaMS]
        else:
            t_DeltaMS = DeltaMS
            t_alpha = np.interp(t_DeltaMS, arr_DeltaMS, arr_alpha)
            t_beta = np.interp(t_DeltaMS, arr_DeltaMS, arr_beta)
    else:
        t_DeltaMS = np.array(DeltaMS)
        t_alpha = t_DeltaMS * 0.0
        t_beta = t_DeltaMS * 0.0
        t_mask_1 = (t_DeltaMS <= min_DeltaMS)
        if np.count_nonzero(t_mask_1) > 0:
            t_DeltaMS[t_mask_1] = min_DeltaMS
            t_alpha[t_mask_1] = arr_alpha[argmin_DeltaMS]
            t_beta[t_mask_1] = arr_beta[argmin_DeltaMS]
        t_mask_2 = (t_DeltaMS >= max_DeltaMS)
        if np.count_nonzero(t_mask_2) > 0:
            t_DeltaMS[t_mask_2] = max_DeltaMS
            t_alpha[t_mask_2] = arr_alpha[argmax_DeltaMS]
            t_beta[t_mask_2] = arr_beta[argmax_DeltaMS]
        t_mask_3 = np.logical_and(t_DeltaMS > min_DeltaMS, t_DeltaMS < max_DeltaMS)
        if np.count_nonzero(t_mask_3) > 0:
            t_alpha[t_mask_3] = np.interp(t_DeltaMS[t_mask_3], arr_DeltaMS, arr_alpha)
            t_beta[t_mask_3] = np.interp(t_DeltaMS[t_mask_3], arr_DeltaMS, arr_beta)
    Mmol = 10**(t_alpha + t_beta * np.log10(SFR))
    return Mmol


def calc_SFR_from_Mmolgas_using_SF_law_Sargent2014(Mmolgas, DeltaMS = 0.0):
    # Sargent et al. 2014, https://ui.adsabs.harvard.edu/abs/2014ApJ...793...19S/abstract
    # bimodal SF law
    arr_alpha = [9.22, 8.05]
    arr_beta = [0.81, 0.81]
    arr_DeltaMS = [0.0, 1.0]
    argmin_DeltaMS = np.argmin(arr_DeltaMS)
    argmax_DeltaMS = np.argmax(arr_DeltaMS)
    min_DeltaMS = arr_DeltaMS[argmin_DeltaMS]
    max_DeltaMS = arr_DeltaMS[argmax_DeltaMS]
    # 
    if np.isscalar(DeltaMS):
        if DeltaMS <= min_DeltaMS:
            t_DeltaMS = min_DeltaMS
            t_alpha = arr_alpha[argmin_DeltaMS]
            t_beta = arr_beta[argmin_DeltaMS]
        elif DeltaMS >= max_DeltaMS:
            t_DeltaMS = max_DeltaMS
            t_alpha = arr_alpha[argmax_DeltaMS]
            t_beta = arr_beta[argmax_DeltaMS]
        else:
            t_DeltaMS = DeltaMS
            t_alpha = np.interp(t_DeltaMS, arr_DeltaMS, arr_alpha)
            t_beta = np.interp(t_DeltaMS, arr_DeltaMS, arr_beta)
    else:
        t_DeltaMS = np.array(DeltaMS)
        t_alpha = t_DeltaMS * 0.0
        t_beta = t_DeltaMS * 0.0
        t_mask_1 = (t_DeltaMS <= min_DeltaMS)
        if np.count_nonzero(t_mask_1) > 0:
            t_DeltaMS[t_mask_1] = min_DeltaMS
            t_alpha[t_mask_1] = arr_alpha[argmin_DeltaMS]
            t_beta[t_mask_1] = arr_beta[argmin_DeltaMS]
        t_mask_2 = (t_DeltaMS >= max_DeltaMS)
        if np.count_nonzero(t_mask_2) > 0:
            t_DeltaMS[t_mask_2] = max_DeltaMS
            t_alpha[t_mask_2] = arr_alpha[argmax_DeltaMS]
            t_beta[t_mask_2] = arr_beta[argmax_DeltaMS]
        t_mask_3 = np.logical_and(t_DeltaMS > min_DeltaMS, t_DeltaMS < max_DeltaMS)
        if np.count_nonzero(t_mask_3) > 0:
            t_alpha[t_mask_3] = np.interp(t_DeltaMS[t_mask_3], arr_DeltaMS, arr_alpha)
            t_beta[t_mask_3] = np.interp(t_DeltaMS[t_mask_3], arr_DeltaMS, arr_beta)
    SFR = 10**( (np.log10(Mmolgas) - t_alpha) / t_beta )
    return SFR








#!/usr/bin/env python
# 

import os, sys
from matplotlib import pyplot as plt
from matplotlib import ticker as ticker
from matplotlib.ticker import MultipleLocator, AutoMinorLocator
from matplotlib import font_manager
import matplotlib as mpl
import numpy as np

def setup_matplotlib():
    
    if os.path.isdir(os.path.expanduser('~')+os.sep+'Library'+os.sep+'Fonts'):
        font_dirs = [os.path.expanduser('~')+os.sep+'Library'+os.sep+'Fonts']
        font_files = font_manager.findSystemFonts(fontpaths=font_dirs)
        font_list = font_manager.createFontList(font_files)
        font_manager.fontManager.ttflist.extend(font_list)
        font_manager.findfont('NGC', rebuild_if_missing=True)
    
    mpl.rcParams['font.family'] = 'NGC'
    
    #mpl.rcParams['text.usetex'] = True
    mpl_version = np.array(mpl.__version__.split('.')).astype(int)
    if mpl_version[0] > 3 or (mpl_version[0] >= 3 and mpl_version[1] >= 1):
        # since matplotlib version 3.1.0, mpl.rcParams['text.latex.preamble'] is a single str not a list
        mpl.rcParams['text.latex.preamble'] = r'\usepackage{amsmath}'
        mpl.rcParams['text.latex.preamble'] += '\n'
        mpl.rcParams['text.latex.preamble'] += r'\makeatletter \newcommand*{\rom}[1]{\expandafter\@slowromancap\romannumeral #1@} \makeatother'
    else:
        # since matplotlib version 3.1.0, mpl.rcParams['text.latex.preamble'] is a single str not a list
        mpl.rcParams['text.latex.preamble'] = [r'\usepackage{amsmath}'] #for \text command
        mpl.rcParams['text.latex.preamble'].append(r'\makeatletter \newcommand*{\rom}[1]{\expandafter\@slowromancap\romannumeral #1@} \makeatother')
    
    mpl.rcParams['axes.labelsize'] = '16' # https://matplotlib.org/users/customizing.html
    mpl.rcParams['axes.grid'] = True
    mpl.rcParams['axes.axisbelow'] = True
    mpl.rcParams['xtick.direction'] = 'in'
    mpl.rcParams['ytick.direction'] = 'in'
    mpl.rcParams['xtick.minor.visible'] = True
    mpl.rcParams['ytick.minor.visible'] = True
    mpl.rcParams['xtick.labelsize'] = '13'
    mpl.rcParams['ytick.labelsize'] = '13'
    mpl.rcParams['xtick.top'] = True
    mpl.rcParams['ytick.right'] = True
    #mpl.rcParams['grid.color'] = 'b0b0b0'
    mpl.rcParams['grid.linestyle'] = '--'
    mpl.rcParams['grid.linewidth'] = 0.25
    mpl.rcParams['grid.alpha'] = 0.8
    mpl.rcParams['legend.fontsize'] = '12'
    mpl.rcParams['legend.borderaxespad'] = 0.2 # space between legend border and axis
    #mpl.rcParams['legend.borderpad'] = 0.2 # space between legend content and legend border
    #mpl.rcParams['legend.handletextpad'] = 0.05
    #mpl.rcParams['legend.labelspacing'] = None # in font-size units
    #mpl.rcParams['legend.columnspacing'] = None # in font-size units
    #mpl.rcParams['legend.ncol']





#!/usr/bin/env python
# 
# usage:
#     from calc_galaxy_size_evolution import (calc_galaxy_size_Tacconi2018,
#                                             )
# 

from __future__ import print_function

import os, sys, re, json, time, astropy
import numpy as np
from astropy.table import Table, Column, hstack
from copy import copy
from numpy import log10, power as pow

if not (os.path.dirname(os.path.abspath(__file__)) in sys.path): sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import apply_cosmology
cosmo = apply_cosmology.cosmo

if sys.version_info.major >= 3:
    long = int
else:
    pass



# 
# def 
# 
def calc_IRX_Whitaker2017(lgMstar, SFR):
    # Whitaker2017 Eq. 1 on page 3
    a = 1.96e9
    b = -2.277
    f_obscured = 1.0 / (1.0 + a * np.exp(b * lgMstar)) # f_obscured = SFR(IR)/SFR(UV+IR)
    # Whitaker2017 Eq. 2 on page 5
    #zlo = np.array([0.02, 0.5, 1.0, 1.5, 2.0])
    #zhi = np.array([0.05, 1.0, 1.5, 2.0, 2.5])
    #a = np.array([1.234, 2.092, 3.917, 6.806, 10.701])
    #b = np.array([-2.858, -2.384, -2.596, -2.673, -2.516])
    #SFR_total = SFR # SFR_total is SFR(UV+IR)
    #f_obscured = 1.0 / (1.0 + a * np.exp(b * np.log10(SFR_total))) # f_obscured = SFR(IR)/SFR(UV+IR)
    # 
    SFR_total = SFR # SFR_total is SFR(UV+IR)
    SFR_IR = SFR_total * f_obscured
    SFR_UV = SFR_total * (1.0-f_obscured)
    IRX = np.log10(SFR_IR/SFR_UV)
    # 
    return IRX, f_obscured, SFR_UV, SFR_IR

def calc_IRX_Schreiber2017(z, lgMstar, SFR):
    # Schreiber2017 Eq. 13 on page 8
    IRX = (0.45 * z + 0.35) * (lgMstar-10.5) + 1.2 # log10(LIR/LUV)
    if np.isscalar(IRX):
        is_scalar = True
    if is_scalar:
        IRX = np.array([IRX])
    # 
    IRX[(z>3.0)] = (0.45 * 3.0 + 0.35) * (lgMstar[(z>3.0)]-10.5) + 1.2 # log10(LIR/LUV)
    if is_scalar:
        IRX = IRX[0]
    # 
    f_obscured = 1.0/(1.0+(1.0/(10**IRX)))
    # 
    SFR_total = SFR
    SFR_IR = SFR_total * 1.0/(1.0+(1.0/(10**IRX)))
    SFR_UV = SFR_total * 1.0/(1.0+(10**IRX))
    IRX = np.log10(SFR_IR/SFR_UV)
    # 
    return IRX, f_obscured, SFR_UV, SFR_IR







